/* 
   In our "high level" application code (as opposed to
   low-level error loggers and other utilities), most
   developers write whatever SQL they need wherever they
   need it.
   
   And, of course, the same or very similar statements
   are needed over and over again, especially for key
   tables.
   
   The result is a proliferation of SQL: hard to optimize,
   very difficult to maintain, quite expensive to upgrade.
*/

CREATE OR REPLACE PROCEDURE do_stuff_with_employee (
   employee_id_in IN employees.employee_id%TYPE)
IS
   l_employee   employees%ROWTYPE;
BEGIN
   SELECT *
     INTO l_employee
     FROM employees
    WHERE employee_id = do_stuff_with_employee.employee_id_in;
    
   /* Do some stuff... */
END;
/

CREATE OR REPLACE PROCEDURE do_stuff_with_manager (
   employee_id_in IN employees.employee_id%TYPE)
IS
   l_manager   employees%ROWTYPE;
BEGIN
   SELECT m.*
     INTO l_manager
     FROM employees e, employees m
    WHERE e.employee_id = do_stuff_with_manager.employee_id_in
      AND m.employee_id = e.manager_id;   
    
   /* Do some stuff... */
END;
/

/* 
   The approach I take is to write the query once
   and hide it behind a function interface, as in:

   [NOTE: the use of RESULT_CACHE only makes sense if employees table is
   updated relatively infrequently!]
*/

CREATE OR REPLACE PACKAGE employees_sql
IS
   FUNCTION onerow (employee_id_in IN employees.employee_id%TYPE)
      RETURN employees%ROWTYPE RESULT_CACHE;

   FUNCTION your_manager (employee_id_in IN employees.employee_id%TYPE)
      RETURN employees%ROWTYPE RESULT_CACHE;
END;
/

CREATE OR REPLACE PACKAGE BODY employees_sql
IS
   FUNCTION onerow (employee_id_in IN employees.employee_id%TYPE)
      RETURN employees%ROWTYPE RESULT_CACHE
   IS
      onerow_rec   employees%ROWTYPE;
   BEGIN
      SELECT *
        INTO onerow_rec
        FROM employees
       WHERE employee_id = employee_id_in;

      RETURN onerow_rec;
   END;

   FUNCTION your_manager (employee_id_in IN employees.employee_id%TYPE)
      RETURN employees%ROWTYPE RESULT_CACHE
   IS
      onerow_rec   employees%ROWTYPE;
   BEGIN
      SELECT m.*
       INTO onerow_rec
       FROM employees e, employees m
      WHERE e.employee_id = onerow.employee_id_in
        AND m.employee_id = e.manager_id; 

      RETURN onerow_rec;
   END;
END;
/

/*
   The result is less code to write and maintain
   (increased productivity!) and a "single point of definition"
   for this particular SQL "service".
   
   And now when I decide I would like result cache queries
   against the employees table, I add RESULT_CACHE in just one
   place (spec and body) and I am done!
*/
      
CREATE OR REPLACE PROCEDURE do_stuff_with_employee (
   employee_id_in IN employees.employee_id%TYPE)
IS
   l_employee   employees%ROWTYPE;
BEGIN
   l_employee := employees_sql.one_row (employee_id_in);
    
   /* Do some stuff... */
END;
/

CREATE OR REPLACE PROCEDURE do_stuff_with_manager (
   employee_id_in IN employees.employee_id%TYPE)
IS
   l_manager   employees%ROWTYPE;
BEGIN
   l_manager := employees_sql.your_manager (employee_id_in);
    
   /* Do some stuff... */
END;
/

/* BUT BUT BUT watch out for "atomizing" your SQL operations: */

CREATE OR REPLACE PROCEDURE do_stuff_with_manager (
   employee_id_in IN employees.employee_id%TYPE)
IS
   l_manager   employees%ROWTYPE;
   l_employee   employees%ROWTYPE;
BEGIN
   /* BAD IDEA! Two SELECTS, two context switches, potential
                read consistency issues! */
   l_employee := employees_sql.one_row (employee_id_in);
   l_manager := employees_sql.your_manager (l_employee.manager_id);
    
   /* Do some stuff... */
END;
/


