CREATE OR REPLACE PACKAGE plscope_hlpr
   AUTHID CURRENT_USER
IS
   PROCEDURE show_dml_statements_on (
      owner_in            IN VARCHAR2,
      table_in            IN VARCHAR2,
      statement_type_in   IN VARCHAR2 DEFAULT 'ALL');

   /* Find all code units that reference the column identifier
   table_in.column_in by first getting the identifier signature
   and using that to find all REFERENCES to the identifier
   that has that signature. */

   PROCEDURE show_column_usages (owner_in    IN VARCHAR2,
                                 table_in    IN VARCHAR2,
                                 column_in   IN VARCHAR2);

   PROCEDURE show_dynamic_sql_usages (owner_in IN VARCHAR2 DEFAULT NULL);

   PROCEDURE full_report_for (owner_in         IN VARCHAR2,
                                      object_name_in   IN VARCHAR2);
END;
/

CREATE OR REPLACE PACKAGE BODY plscope_hlpr
IS
   PROCEDURE show_dml_statements_on (
      owner_in            IN VARCHAR2,
      table_in            IN VARCHAR2,
      statement_type_in   IN VARCHAR2 DEFAULT 'ALL')
   IS
   BEGIN
      DBMS_OUTPUT.put_line (
            'Locations of '
         || statement_type_in
         || ' DML Statements Against '
         || owner_in
         || '.'
         || table_in);
      DBMS_OUTPUT.put_line ('');

      FOR rec
         IN (  SELECT idt.line,
                      idt.owner || '.' || idt.object_name code_unit,
                      RTRIM (src.text, CHR (10)) text
                 FROM all_identifiers idt, all_source src, all_statements st
                WHERE     idt.usage = 'REFERENCE'
                      AND idt.TYPE = 'TABLE'
                      AND idt.name = table_in
                      AND idt.owner = owner_in
                      AND idt.line = src.line
                      AND idt.object_name = src.name
                      AND idt.owner = src.owner
                      AND idt.usage_context_id = st.usage_id
                      AND (   st.TYPE = statement_type_in
                           OR statement_type_in = 'ALL')
             ORDER BY st.TYPE,
                      idt.owner,
                      idt.object_name,
                      idt.line)
      LOOP
         DBMS_OUTPUT.put_line (
            'In ' || rec.code_unit || ' on ' || rec.line || ': ' || rec.text);
      END LOOP;
   END;

   PROCEDURE show_column_usages (owner_in    IN VARCHAR2,
                                 table_in    IN VARCHAR2,
                                 column_in   IN VARCHAR2)
   IS
   BEGIN
      DBMS_OUTPUT.put_line (
         'References to ' || owner_in || '.' || table_in || '.' || column_in);
      DBMS_OUTPUT.put_line ('');

      FOR rec
         IN (  SELECT idt.line,
                      idt.owner || '.' || idt.object_name code_unit,
                      RTRIM (src.text, CHR (10)) text
                 FROM all_identifiers idt, all_source src
                WHERE     idt.usage = 'REFERENCE'
                      AND idt.signature =
                             (SELECT idt_inner.signature
                                FROM all_identifiers idt_inner
                               WHERE     idt_inner.object_name =
                                            show_column_usages.table_in
                                     AND idt_inner.TYPE = 'COLUMN'
                                     AND idt_inner.name =
                                            show_column_usages.column_in
                                     AND idt_inner.usage = 'DECLARATION'
                                     AND idt_inner.owner = owner_in)
                      AND idt.line = src.line
                      AND idt.object_name = src.name
                      AND idt.owner = src.owner
             ORDER BY code_unit, line)
      LOOP
         DBMS_OUTPUT.put_line (
            'In ' || rec.code_unit || ' on ' || rec.line || ': ' || rec.text);
      END LOOP;
   END;

   PROCEDURE show_dynamic_sql_usages (owner_in IN VARCHAR2 DEFAULT NULL)
   IS
   BEGIN
      FOR rec
         IN (SELECT st.owner,
                    st.object_name,
                    st.line,
                    s.text
               FROM all_statements st, all_source s
              WHERE     st.TYPE IN ('EXECUTE IMMEDIATE', 'OPEN')
                    AND st.owner = s.owner
                    AND st.object_name = s.name
                    AND st.line = s.line
             UNION ALL
             SELECT idnt.owner,
                    idnt.object_name,
                    idnt.line,
                    src.text
               FROM all_identifiers idnt, all_source src
              WHERE     idnt.owner <> 'SYS'
                    AND idnt.signature IN (SELECT a.signature
                                             FROM all_identifiers a
                                            WHERE     a.usage = 'DECLARATION'
                                                  AND a.owner = 'SYS'
                                                  AND a.object_name =
                                                         'DBMS_SQL'
                                                  AND a.object_type =
                                                         'PACKAGE')
                    AND idnt.owner = src.owner
                    AND idnt.name = src.name
                    AND idnt.line = src.line)
      LOOP
         NULL;
      END LOOP;
   END;

   PROCEDURE full_report_for (owner_in         IN VARCHAR2,
                                      object_name_in   IN VARCHAR2)
   IS
   BEGIN
      FOR rec
         IN (    SELECT plscope_type,
                        usage_id,
                        usage_context_id,
                        LPAD (' ', 2 * (LEVEL - 1)) || usage || ' ' || name
                           usages,
                        line,
                        col,
                        signature
                   FROM (SELECT 'ID' plscope_type,
                                ai.object_name,
                                ai.usage usage,
                                ai.usage_id,
                                ai.usage_context_id,
                                ai.TYPE || ' ' || ai.name name,
                                ai.line,
                                ai.col,
                                signature
                           FROM all_identifiers ai
                          WHERE     ai.owner = owner_in
                                AND ai.object_name = object_name_in
                         UNION ALL
                         SELECT 'ST',
                                st.object_name,
                                st.TYPE,
                                st.usage_id,
                                st.usage_context_id,
                                'STATEMENT',
                                st.line,
                                st.col,
                                signature
                           FROM all_statements st
                          WHERE     st.owner = owner_in
                                AND st.object_name = object_name_in)
             START WITH usage_context_id = 0
             CONNECT BY PRIOR usage_id = usage_context_id)
      LOOP
         NULL;
      END LOOP;
   END;
END;
/