/*
Aggregates and their necessary adjunct, qualified expressions, improve program clarity and programmer productivity.

Through Oracle Database 12c release 2, it was possible to supply the value of a non-scalar datatype with an expression, 
for example by using the type constructor for a nested table. Starting with Oracle Database Release 18c, any PL/SQL 
value can be provided by an expression (for example for a record or for an associative array) like a constructor 
provides an abstract datatype value. In PL/SQL, we use the terms "qualified expression" and "aggregate" rather 
than the SQL term "type constructor", but the functionality is the same.

Qualified expressions improve program clarity and developer productivity by providing the ability to declare and 
define a complex value in a compact form where the value is needed.

A qualified expression combines expression elements to create values of a RECORD type or associative array type.

Qualified expressions use an explicit type indication to provide the type of the qualified item. This explicit 
indication is known as a typemark.
*/

DECLARE
   TYPE species_rt IS RECORD
   (
      species_name           VARCHAR2 (100),
      habitat_type           VARCHAR2 (100),
      surviving_population   INTEGER
   );

   l_elephant   species_rt := species_rt ('Elephant', 'Savannah', '10000');

   PROCEDURE display_species (
      species_in    species_rt DEFAULT species_rt ('Not Set', 'Global', 0))
   IS
   BEGIN
      DBMS_OUTPUT.put_line ('Species: ' || species_in.species_name);
      DBMS_OUTPUT.put_line ('Habitat: ' || species_in.habitat_type);
      DBMS_OUTPUT.put_line ('# Left: ' || species_in.surviving_population);
   END;
BEGIN
   display_species (l_elephant);

   /* Use the default */
   display_species ();
END;
/

/* But not for %ROWTYPE records :-( 

PLS-00431: bulk SQL attributes must use a single index

*/

create table species 
   (
      species_name           VARCHAR2 (100),
      habitat_type           VARCHAR2 (100),
      surviving_population   INTEGER
)
/

DECLARE
   l_elephant   species%rowtype := species%rowtype ('Elephant', 'Savannah', '10000');
BEGIN
   NULL;
END;
/


/* With named notation */


DECLARE
   TYPE species_rt IS RECORD
   (
      species_name           VARCHAR2 (100),
      habitat_type           VARCHAR2 (100),
      surviving_population   INTEGER
   );

   l_elephant   species_rt
      := species_rt (species_name           => 'Elephant',
                     surviving_population   => '10000',
                     habitat_type           => 'Savannah');

   PROCEDURE display_species (
      species_in    species_rt DEFAULT species_rt ('Not Set', 'Global', 0))
   IS
   BEGIN
      DBMS_OUTPUT.put_line ('Species: ' || species_in.species_name);
      DBMS_OUTPUT.put_line ('Habitat: ' || species_in.habitat_type);
      DBMS_OUTPUT.put_line ('# Left: ' || species_in.surviving_population);
   END;
BEGIN
   display_species (l_elephant);

   /* Use the default */
   display_species ();
END;
/


DECLARE
   TYPE species1_rt IS RECORD
   (
      species_name           VARCHAR2 (100),
      habitat_type           VARCHAR2 (100),
      surviving_population   INTEGER
   );

   TYPE species2_rt IS RECORD
   (
      species_name           VARCHAR2 (100),
      habitat_type           VARCHAR2 (100),
      surviving_population   INTEGER
   );

   l_species1   species1_rt := species1_rt ('Elephant', 'Savannah', '10000');

   l_species2   species2_rt;
BEGIN
   l_species2 := l_species1;
END;
/

DECLARE
   TYPE species1_rt IS RECORD
   (
      species_name           VARCHAR2 (100),
      habitat_type           VARCHAR2 (100),
      surviving_population   INTEGER
   );

   TYPE species2_rt IS RECORD
   (
      species_name           VARCHAR2 (100),
      habitat_type           VARCHAR2 (100),
      surviving_population   INTEGER
   );

   l_species1   species1_rt := species1_rt ('Elephant', 'Savannah', '10000');

   l_species2   species2_rt;
BEGIN
   l_species2 :=
      species2_rt (l_species1.species_name,
                   l_species1.habitat_type,
                   l_species1.surviving_population);
END;
/