/* Clean start */

DECLARE
   l_status   PLS_INTEGER;
BEGIN
   l_status := dbms_soda.drop_collection ('FriendsWithDocs');
   l_status := dbms_soda.drop_collection ('MyOwnKeys');
   l_status := dbms_soda.drop_collection ('ColorKeys');
   l_status := dbms_soda.drop_collection ('SillyCollection');
   COMMIT;
END;
/

/* Create a collection for use in these scripts */

DECLARE
   l_collection   soda_collection_t;
   l_status         PLS_INTEGER;
BEGIN
   l_collection := dbms_soda.create_collection ('FriendsWithDocs');

   l_status :=
      l_collection.insert_one (
         soda_document_t (
            b_content   => UTL_RAW.cast_to_raw (
                             '{"friend_type":1,
                               "friend_name":"Lakshmi", 
                               "favorites": [{"song" : "Somewhere Over the Rainbow"},
                                             {"spice" : "tamarind"},
                                             {"flavor" : "cherry"} ]}')));

   l_status :=
      l_collection.insert_one (
         soda_document_t (
            b_content   => UTL_RAW.cast_to_raw (
                             '{"friend_type":2,
                               "friend_name":"Sally", 
                               "favorites": [{"color" : "blue"},
                                             {"flavor" : "chocolate"},
                                             {"flower" : "rose"} ]}')));

   l_status :=
      l_collection.insert_one (
         soda_document_t (
            b_content   => UTL_RAW.cast_to_raw (
                             '{"friend_type":2,
                               "friend_name":"Jorge", 
                               "favorites": [{"color" : "green"},
                                             {"flavor" : "chocolate"},
                                             {"tree" : "oak"} ]}')));
END;
/

/* Chaining */

DECLARE
   l_collection   soda_collection_t;
   l_operation    soda_operation_t;
   l_cursor       soda_cursor_t;
   l_status         BOOLEAN;
BEGIN
   l_collection := dbms_soda.open_collection ('FriendsWithDocs');

   l_operation := l_collection.find ();

   l_cursor := l_operation.get_cursor ();

   l_status := l_cursor.close;
END;
/

DECLARE
   l_cursor   soda_cursor_t;
   l_status     BOOLEAN;
BEGIN
   l_cursor :=
      dbms_soda.open_collection ('FriendsWithDocs').find ().get_cursor ();

   l_status := l_cursor.close;
END;
/

/* Iterate through all documents using the keys */

DECLARE
   l_document   soda_document_t;
   l_cursor     soda_cursor_t;
   l_content_shown BOOLEAN := FALSE;
   l_status       BOOLEAN;
BEGIN
   l_cursor :=
      dbms_soda.open_collection ('FriendsWithDocs').find ().get_cursor ();

   -- Loop through the cursor
   WHILE l_cursor.has_next
   LOOP
      l_document := l_cursor.NEXT;

      IF l_document IS NOT NULL
      THEN
         DBMS_OUTPUT.put_line ('Document key: ' || l_document.get_key);
         
         IF NOT l_content_shown
         THEN
            l_content_shown := TRUE;
            DBMS_OUTPUT.put_line (
               'Content: ' || json_query (l_document.get_blob, '$' PRETTY));
         END IF;
         
         DBMS_OUTPUT.put_line('Creation timestamp: ' || l_document.get_created_on);
         DBMS_OUTPUT.put_line('Last modified timestamp: ' || l_document.get_last_modified);
         DBMS_OUTPUT.put_line('Version: ' || l_document.get_version);
      END IF;
    END LOOP;

    l_status := l_cursor.close;
END;
/

/* Iterate through a filtered list of documents using JSON l_filter syntax. */

DECLARE
   l_document   soda_document_t;
   l_cursor     soda_cursor_t;
   l_filter          VARCHAR2 (128) := '{"friend_type" : "2"}';
   l_status       BOOLEAN;
BEGIN
   l_cursor :=
      dbms_soda.open_collection ('FriendsWithDocs').find ().filter (l_filter).get_cursor ();

   WHILE l_cursor.has_next
   LOOP
      l_document := l_cursor.NEXT;

      IF l_document IS NOT NULL
      THEN
         DBMS_OUTPUT.put_line ('Document key: ' || l_document.get_key);
         DBMS_OUTPUT.put_line (
            'Content: ' || json_query (l_document.get_blob, '$' PRETTY));
      END IF;
    END LOOP;

    l_status := l_cursor.close;
END;
/

DECLARE
   l_document   soda_document_t;
   l_cursor     soda_cursor_t;
   l_filter          VARCHAR2 (128) := '{"favorites.flavor" : "chocolate"}';
   l_status       BOOLEAN;
BEGIN
   l_cursor :=
      dbms_soda.open_collection ('FriendsWithDocs').find ().filter (l_filter).get_cursor ();

   WHILE l_cursor.has_next
   LOOP
      l_document := l_cursor.NEXT;

      IF l_document IS NOT NULL
      THEN
         DBMS_OUTPUT.put_line ('Document key: ' || l_document.get_key);
         DBMS_OUTPUT.put_line (
            'Content: ' || json_query (l_document.get_blob, '$' PRETTY));
      END IF;
    END LOOP;

    l_status := l_cursor.close;
END;
/

/* Find a document with a specific key */

DECLARE
   l_document     soda_document_t;
   l_collection   soda_collection_t;
   l_metadata       VARCHAR2 (4000)
      := '{"keyColumn" : {"name" : "KEY", "assignmentMethod" : "CLIENT" }}';
   l_status         PLS_INTEGER;
BEGIN
   l_collection := dbms_soda.create_collection ('MyOwnKeys', l_metadata);
   
   DBMS_OUTPUT.put_line (
         'Collection specification: '
      || json_query (l_collection.get_metadata, '$' PRETTY));
      
   l_status :=
      l_collection.insert_one (
         soda_document_t ('FriendLakshmi',
            b_content   => UTL_RAW.cast_to_raw (
                             '{"friend_type":1,"friend_name":"Lakshmi"}')));

   l_document :=
      l_collection
         .find ()
         .key ('FriendLakshmi')
         .get_one;

   DBMS_OUTPUT.put_line ('Key: ' || l_document.get_key);
   DBMS_OUTPUT.put_line ('Content: ');
   DBMS_OUTPUT.put_line (   
      l_document.get_key|| '+'||json_query (l_document.get_blob, '$' PRETTY));
END;
/

/* Find documents with list of keys */

DECLARE
   l_collection   soda_collection_t;
   l_metadata     VARCHAR2 (4000)
      := '{"keyColumn" : {"name" : "KEY",  "assignmentMethod": "CLIENT" }}';
   l_keys         soda_key_list_t;
   l_cursor       soda_cursor_t;
   l_status       BOOLEAN;
   l_document     soda_document_t;
BEGIN
   l_collection := dbms_soda.create_collection ('ColorKeys', l_metadata);

   l_status :=
      l_collection.insert_one (
         soda_document_t (
            'Red',
            b_content   => UTL_RAW.cast_to_raw ('{"thing":"blood"}'))) = 1;
   l_status :=
      l_collection.insert_one (
         soda_document_t (
            'Green',
            b_content   => UTL_RAW.cast_to_raw ('{"thing":"grass"}'))) = 1;

   l_keys := soda_key_list_t ('Red', 'Green');

   l_cursor := l_collection.find ().keys (l_keys).get_cursor ();

   WHILE l_cursor.has_next
   LOOP
      l_document := l_cursor.NEXT;

      IF l_document IS NOT NULL
      THEN
         DBMS_OUTPUT.put_line ('Key: ' || l_document.get_key);
         DBMS_OUTPUT.put_line (
            json_query (l_document.get_blob, '$' PRETTY));
      END IF;
    END LOOP;

    l_status := l_cursor.close;
END;
/

/*
Key: Green
{
  "thing" : "grass"
}
Key: Red
{
  "thing" : "blood"
}
*/

/* Replace a document */

DECLARE
    l_collection  SODA_COLLECTION_T;
    l_document    SODA_DOCUMENT_T;
    l_new_document SODA_DOCUMENT_T;
BEGIN
    l_collection := DBMS_SODA.open_collection('ColorKeys');
    l_document := SODA_DOCUMENT_T(
                  b_content => utl_raw.cast_to_raw('{"Blue" : "Bluebird"}'));
    l_new_document := l_collection.find().key('Red').replace_one_and_get(l_document);

    IF l_new_document IS NOT NULL 
    THEN
        DBMS_OUTPUT.put_line('Document:');
        DBMS_OUTPUT.put_line('Key: ' || l_new_document.get_key);
        DBMS_OUTPUT.put_line('Creation timestamp: ' || l_new_document.get_created_on);
    END IF;
END;
/

/* Remove a document */

DECLARE
    l_collection  SODA_COLLECTION_T;
    l_status NUMBER;
BEGIN
    l_collection := DBMS_SODA.open_collection('ColorKeys');
    
    l_status:= l_collection.find().key('Green').remove;

    IF  l_status = 1 
    THEN
        DBMS_OUTPUT.put_line('Document removed');
    END IF;
END; 
/

/* Pagination built right into the API 

Note: ordering is by the internal key, so it appears to be random.

*/

DECLARE
   l_collection   soda_collection_t;
   l_document     soda_document_t;
   l_cursor       soda_cursor_t;
   l_status       BOOLEAN;
BEGIN
   l_collection := dbms_soda.create_collection ('SillyCollection');

   FOR indx IN 1 .. 100
   LOOP
      l_status :=
         l_collection.insert_one (
            soda_document_t (
               b_content   => UTL_RAW.cast_to_raw (
                                   '{"myIndex":'
                                || indx
                                || ',"myValue":"'
                                || CASE MOD (indx, 2)
                                      WHEN 0 THEN 'Even'
                                      ELSE 'Odd'
                                   END
                                || '"}'))) = 1;
   END LOOP;

   l_cursor := l_collection.find ().skip (50).LIMIT (5).get_cursor;

   WHILE l_cursor.has_next
   LOOP
      l_document := l_cursor.NEXT;

      IF l_document IS NOT NULL
      THEN
         DBMS_OUTPUT.put_line (
            'Content: ' || json_query (l_document.get_blob, '$' PRETTY));
      END IF;
    END LOOP;

    l_status := l_cursor.close;
END;
/

/* Sort using QBE */

DECLARE
   l_collection   soda_collection_t;
   l_document     soda_document_t;
   l_cursor       soda_cursor_t;
   l_filter VARCHAR2(100);
   l_status       BOOLEAN;
BEGIN
   l_collection := dbms_soda.open_collection ('SillyCollection');

   l_filter := '{"$orderby" : { "myIndex" : 1}}';
   l_cursor := l_collection.find().filter(l_filter).skip(50).limit(5).get_cursor();

   WHILE l_cursor.has_next
   LOOP
      l_document := l_cursor.NEXT;

      IF l_document IS NOT NULL
      THEN
         DBMS_OUTPUT.put_line (
            'Content: ' || json_query (l_document.get_blob, '$' PRETTY));
      END IF;
    END LOOP;

    l_status := l_cursor.close;
END;
/