CREATE OR REPLACE PACKAGE alert_mgr
   AUTHID DEFINER
/* Copyright 2020 Oracle Corporation */
/*
Prototype for utility to enable triggering of actions based
on a change in data that is defined by an alert.

Dev Gym example:

If no one has answered a quiz in last hour, something
may be wrong with the site, so notify admins.

The alert is registered (row put in alerts table).

A job runs every 5 minutes and checks to see if any active
alerts are due to be fired.

If so, run the trigger code to see if the action should be
triggered. If not, report in alert_runs table. If so, execute
the code as specified and store results in alert_runs table.

*/
IS
   c_job_name                        CONSTANT VARCHAR2 (100) := 'CHECK_ALERTS_JOB';

   /* Queries that return 1 or more rows return a value that can be converted to a string */
   c_trigger_type_select_one_row     CONSTANT VARCHAR2 (100) := 'ONEROW';
   c_trigger_type_select_no_rows     CONSTANT VARCHAR2 (100) := 'NOROWS';
   c_trigger_type_select_mult_rows   CONSTANT VARCHAR2 (100) := 'MULTROWS';
   c_trigger_type_return_true        CONSTANT VARCHAR2 (100) := 'RETURNTRUE';
   c_trigger_type_return_false       CONSTANT VARCHAR2 (100) := 'RETURNFALSE';

   c_action_type_email               CONSTANT VARCHAR2 (100) := 'EMAIL';
   c_action_type_sms                 CONSTANT VARCHAR2 (100) := 'SMS';
   c_action_type_proc                CONSTANT VARCHAR2 (100) := 'PROC';
   c_action_type_rest                CONSTANT VARCHAR2 (100) := 'REST';

   c_freq_hourly                     CONSTANT VARCHAR2 (1) := 'H';
   c_freq_daily                      CONSTANT VARCHAR2 (1) := 'D';

   /* utilities */

   PROCEDURE settrc (on_in IN BOOLEAN, send_to_in IN VARCHAR2 DEFAULT 'SCREEN');

   PROCEDURE trc (val IN VARCHAR2);

   PROCEDURE trc (val IN BOOLEAN);

   PROCEDURE trc (str IN VARCHAR2, val IN BOOLEAN);

   PROCEDURE LOG (info_in IN VARCHAR2);

   --
   PROCEDURE pl (val IN VARCHAR2);

   PROCEDURE bpl (val IN BOOLEAN);

   PROCEDURE bpl (str IN VARCHAR2, val IN BOOLEAN);

   /* job management */

   /* Run once to create the job and get it going */
   PROCEDURE start_check_alerts_job (every_n_minutes_in IN NUMBER DEFAULT 5);

   /* Change the frequencey with which alerts are checke */
   PROCEDURE change_check_frequency (every_n_minutes_in IN NUMBER);

   /* Job is still running, but it doesn't check anything */
   PROCEDURE stop_checking;

   PROCEDURE resume_checking;

   FUNCTION checking_for_alerts
      RETURN BOOLEAN;

   /* Not yet implemented 
   FUNCTION check_alerts_job_running
      RETURN BOOLEAN;

   FUNCTION check_alerts_job_running_yn
      RETURN VARCHAR2;

   PROCEDURE check_alerts_job_status;
   */

   /* Specify frequency as:
      # - number of minutes after which to check
      H - Once an hour = #60
      D - Once a day = #60X24 */

   PROCEDURE register_alert (
      alert_name_in             IN     VARCHAR2,
      description_in            IN     VARCHAR2 DEFAULT NULL,
      trigger_code_in           IN     VARCHAR2,
      trigger_type_in           IN     VARCHAR2,
      check_frequency_in        IN     VARCHAR2,
      recurrence_frequency_in   IN     VARCHAR2,
      ignore_failures_in        IN     VARCHAR2,
      action_type_in            IN     VARCHAR2,
      action_code_in            IN     VARCHAR2,
      action_email_in           IN     VARCHAR2 DEFAULT NULL,
      is_active_in              IN     VARCHAR2 DEFAULT 'N',
      alert_id_out                 OUT NUMBER);

   PROCEDURE register_alert (alert_name_in             IN VARCHAR2,
                             description_in            IN VARCHAR2 DEFAULT NULL,
                             trigger_code_in           IN VARCHAR2,
                             trigger_type_in           IN VARCHAR2,
                             check_frequency_in        IN VARCHAR2,
                             recurrence_frequency_in   IN VARCHAR2,
                             ignore_failures_in        IN VARCHAR2,
                             action_type_in            IN VARCHAR2,
                             action_code_in            IN VARCHAR2,
                             action_email_in           IN VARCHAR2 DEFAULT NULL,
                             is_active_in              IN VARCHAR2 DEFAULT 'N');

   /* Activate an alert for checking *
   PROCEDURE activate_alert (id_in IN NUMBER);

   /* Ignore this alert when checking */
   PROCEDURE deactivate_alert (id_in IN NUMBER);

   /* Run by job at specified frequency to see if any
    alerts have been triggered */
   PROCEDURE check_alerts;

   FUNCTION time_to_trigger_alert (check_frequency_in          IN VARCHAR2,
                                   recurring_frequency_in      IN VARCHAR2,
                                   completed_at_in             IN DATE,
                                   completed_successfully_in   IN VARCHAR2,
                                   ignore_failures_in          IN VARCHAR2)
      RETURN BOOLEAN;

   PROCEDURE execute_alert_action (alert_id_in IN NUMBER);

   /* Add to the alert runs tables and also update the alert row with
      the last run id */
   PROCEDURE record_result (
      alert_id_in                 IN NUMBER,
      completed_at_in             IN DATE DEFAULT SYSDATE,
      completed_successfully_in   IN VARCHAR2 DEFAULT 'Y',
      results_in                  IN VARCHAR2 DEFAULT NULL,
      errors_in                   IN VARCHAR2 DEFAULT NULL);

   FUNCTION one_alert (id_in IN NUMBER)
      RETURN alerts%ROWTYPE;

   FUNCTION last_run_for (alert_id_in IN NUMBER)
      RETURN alert_runs%ROWTYPE;
END;
/