DROP TABLE alert_log
/

DROP TABLE alert_runs
/

DROP TABLE alerts
/

DROP TABLE alert_config
/

CREATE TABLE alert_config (
   job_wakeup_minutes INTEGER,
   keep_checking VARCHAR2(1))
/

begin insert into alert_config values (5, 'Y'); commit; end;
/

CREATE TABLE alerts
(
   id                     NUMBER NOT NULL CONSTRAINT alerts_id_pk PRIMARY KEY,
   alert_name             VARCHAR2 (100) CONSTRAINT alerts_alert_name_unq UNIQUE,
   description            VARCHAR2 (4000),
   trigger_code           VARCHAR2 (4000),
   trigger_type           VARCHAR2 (100),
   check_frequency        VARCHAR2 (1000),
   recurrence_frequency   VARCHAR2 (1000),
   ignore_failures        VARCHAR2 (1),
   action_type            VARCHAR2 (100),
   action_code            VARCHAR2 (4000),
   action_email            VARCHAR2 (4000),
   is_active              VARCHAR2 (1),
   last_run_id            INTEGER,
   created                DATE NOT NULL,
   created_by             VARCHAR2 (255) NOT NULL,
   updated                DATE NOT NULL,
   updated_by             VARCHAR2 (255) NOT NULL,
   CONSTRAINT trigger_type_choices CHECK
      (trigger_type IN ('ONEROW',
                        'NOROWS',
                        'MULTROWS',
                        'RETURNTRUE',
                        'RETURNFALSE')),
   CONSTRAINT action_type_choices CHECK
      (action_type IN ('EMAIL',
                       'SMS',
                       'PROC',
                       'REST')),
   CONSTRAINT ignore_failures_yn CHECK (ignore_failures IN ('Y', 'N')),
   CONSTRAINT active_yn CHECK (is_active IN ('Y', 'N'))
)
/

COMMENT ON COLUMN alerts.trigger_code 
   IS 'Code executed to determine if an alert should be triggered'
/

COMMENT ON COLUMN alerts.trigger_type 
   IS 'Type of code executed to determine if an alert should be triggered. See trigger_type_choices constraint.'
/

COMMENT ON COLUMN alerts.check_frequency 
   IS 'How often to check to see if alert is triggered. H for hourly, D for daily, # for number of minutes since last triggering.'
/

COMMENT ON COLUMN alerts.recurrence_frequency 
   IS 'How often to trigger an alert since last triggering. H for hourly, D for daily, # for number of minutes since last triggering.'
/   

COMMENT ON COLUMN alerts.ignore_failures 
   IS 'If Y then failure of last triggered action does not count for recurrence frequency.'
/

COMMENT ON COLUMN alerts.action_type 
   IS 'The type of action taken. See action_type_choices for options.'
/

COMMENT ON COLUMN alerts.action_code 
   IS 'Code used to execute action.'
/

COMMENT ON COLUMN alerts.action_email 
   IS 'Comma-delimited list of email addresses if action is to send an email'
/

COMMENT ON COLUMN alerts.is_active 
   IS 'If Y then include in alert checking, otherwise ignore.'
/

COMMENT ON COLUMN alerts.last_run_id 
   IS 'Primary key of row in alert_runs table which is the last run for this alert.'
/

CREATE TABLE alert_runs
(
    id                       NUMBER NOT NULL CONSTRAINT alert_runs_id_pk PRIMARY KEY,
    alert_id                 NUMBER
                               CONSTRAINT alert_runs_alert_id_fk
                                   REFERENCES alerts ON DELETE CASCADE,
   started_at               DATE,
   completed_at             DATE,
   completed_successfully   VARCHAR2 (1),
   results                  VARCHAR2 (4000),
   errors                   VARCHAR2 (4000),
   created                  DATE NOT NULL,
   created_by               VARCHAR2 (255) NOT NULL,
   updated                  DATE NOT NULL,
   updated_by               VARCHAR2 (255) NOT NULL,
   CONSTRAINT success_yn CHECK (completed_successfully IN ('Y', 'N'))
)
/

CREATE TABLE alert_log
(
    id                       NUMBER NOT NULL CONSTRAINT alert_log_id_pk PRIMARY KEY,
    info varchar2(4000),
    created                  DATE NOT NULL,
   created_by               VARCHAR2 (255) NOT NULL,
   updated                  DATE NOT NULL,
   updated_by               VARCHAR2 (255) NOT NULL
)
/

CREATE OR REPLACE TRIGGER alerts_biu
   BEFORE INSERT OR UPDATE
   ON alerts
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   :new.alert_name := UPPER (:new.alert_name);
   :new.trigger_type := UPPER (:new.trigger_type);
   :new.action_type := UPPER (:new.action_type);
END alerts_biu;
/

CREATE OR REPLACE TRIGGER alert_runs_biu
   BEFORE INSERT OR UPDATE
   ON alert_runs
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
END alert_runs_biu;
/

CREATE OR REPLACE TRIGGER alert_log_biu
   BEFORE INSERT OR UPDATE
   ON alert_log
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
END alert_log_biu;
/  

CREATE INDEX alert_runs_i1
   ON alert_runs (alert_id)
/   

-- load data
update dg_classes set meeting_id=510462941 where class_id=

-- Generated by Quick SQL Friday January 24, 2020  15:39:13

/*
alerts /auditcols
 id
 alert_name vc100 /unique /upper 
 description vc4000
 trigger_code vc4000
 trigger_type vc100 /upper /values ONEROW,NOROWS,MULTROWS,RETURNTRUE,RETURNFALSE
 check_frequency vc1000
 recurrence_frequency vc1000
 action_type vc100 /upper /values EMAIL,SMS,PROC,REST
 
 
alert_runs /auditcols
 id
 alert_id /fk alert
 started_at date
 completed_at date
 successful vc1
 results vc4000
 errors vc4000

# settings = { language: "EN", APEX: true }
*/