CREATE OR REPLACE PACKAGE alert_mgr
   AUTHID DEFINER
/*
Prototype utility to enable triggering of actions based
on a change in data that is defined by an alert.

Author: Steven Feuerstein
Copyright 2020 Oracle

v 0.8

Add support for parameterization

*/
IS
   c_job_name                    CONSTANT VARCHAR2 (100) := 'CHECK_ALERTS_JOB';

   /* ALERT TYPES */

   /* Simple alert types */
   c_trigtype_select_one_row     CONSTANT VARCHAR2 (100) := 'ONEROW';
   c_trigtype_select_no_rows     CONSTANT VARCHAR2 (100) := 'NOROWS';
   c_trigtype_select_mult_rows   CONSTANT VARCHAR2 (100) := 'MULTROWS';
   c_trigtype_return_true        CONSTANT VARCHAR2 (100) := 'RETURNTRUE';
   c_trigtype_return_false       CONSTANT VARCHAR2 (100) := 'RETURNFALSE';

   /* ACTION TYPES */

   c_action_type_proc            CONSTANT VARCHAR2 (100) := 'PROC';
   c_action_type_email           CONSTANT VARCHAR2 (100) := 'EMAIL';

   /* Not currently supported */
   c_action_type_rest            CONSTANT VARCHAR2 (100) := 'REST';
   c_action_type_sms             CONSTANT VARCHAR2 (100) := 'SMS';

   /* Frequencies: you specify one of these three or an integer value, which
        means that number of minutes */
   c_freq_hourly                 CONSTANT VARCHAR2 (1) := 'H';
   c_freq_daily                  CONSTANT VARCHAR2 (1) := 'D';
   c_freq_never                  CONSTANT VARCHAR2 (1) := 'N';

   /* Tags to avoid hard-coding of alert attributes 
   
   When you add a new tag, update replaced_text function.
   */

   c_bind_name_tag               CONSTANT VARCHAR2 (100) := '#BIND_NAME#';
   c_bind_value_tag              CONSTANT VARCHAR2 (100) := '#BIND_VALUE#';
   c_bind_datatype_tag               CONSTANT VARCHAR2 (100) := '#BIND_DATATYPE#';
   c_bind_description_tag        CONSTANT VARCHAR2 (100)
                                             := '#BIND_DESCRIPTION#' ;
   c_key_name_tag                CONSTANT VARCHAR2 (100) := '#KEY_NAME#';
   c_key_value_tag               CONSTANT VARCHAR2 (100) := '#KEY_VALUE#';
   c_check_frequency_tag         CONSTANT VARCHAR2 (100)
                                                := '#CHECK_FREQ_MINUTES#' ;
   c_check_frequency_text_tag         CONSTANT VARCHAR2 (100)
                                                := '#CHECK_FREQ_TEXT#' ;
   c_recur_frequency_tag         CONSTANT VARCHAR2 (100)
                                             := '#RECUR_FREQ_MINUTES#' ;
   c_last_run_time_tag         CONSTANT VARCHAR2 (100)
                                             := '#LAST_RUN_TIME#' ;
   c_recipient_info_tag         CONSTANT VARCHAR2 (100)
                                             := '#RECIPIENT_INFO#' ;                                          

   /* Datatypes used in call-outs */
   TYPE action_recipients_t IS TABLE OF VARCHAR2 (1000)
      INDEX BY PLS_INTEGER;

   TYPE action_rt IS RECORD
   (
      action_sequence   alert_actions.action_sequence%TYPE,
      action_type       alert_actions.action_type%TYPE,
      action_code       alert_actions.action_code%TYPE,
      send_email_to     alert_actions.send_email_to%TYPE,
      email_subject     alert_actions.email_subject%TYPE,
      email_body        alert_actions.email_body%TYPE,
      email_query       alert_actions.email_query%TYPE
   );

   TYPE actions_t IS TABLE OF action_rt
      INDEX BY PLS_INTEGER;

   /* utilities */

   PROCEDURE settrc (on_in IN BOOLEAN, send_to_in IN VARCHAR2 DEFAULT 'SCREEN');

   PROCEDURE trc (val IN VARCHAR2);

   PROCEDURE trc (val IN BOOLEAN);

   PROCEDURE trc (str IN VARCHAR2, val IN BOOLEAN);

   PROCEDURE log_info (info_in IN VARCHAR2);

   PROCEDURE pl (val IN VARCHAR2);

   PROCEDURE bpl (val IN BOOLEAN);

   PROCEDURE bpl (str IN VARCHAR2, val IN BOOLEAN);

   /* job management */

   /* Run once to create the job and get it going */
   PROCEDURE start_check_alerts_job (every_n_minutes_in IN NUMBER DEFAULT 5);

   /* Change the frequencey with which alerts are checke */
   PROCEDURE change_check_frequency (every_n_minutes_in IN NUMBER);

   /* Register email procedure to avoid any external dependencies*/
   PROCEDURE register_email_proc (proc_name_in        IN VARCHAR2,
                                  workspace_name_in   IN VARCHAR2 DEFAULT NULL);

   /* Job is still running, but it doesn't check anything */
   PROCEDURE stop_checking;

   PROCEDURE resume_checking;

   FUNCTION checking_for_alerts
      RETURN BOOLEAN;

   /* Not yet implemented
   FUNCTION check_alerts_job_running
      RETURN BOOLEAN;

   FUNCTION check_alerts_job_running_yn
      RETURN VARCHAR2;

   PROCEDURE check_alerts_job_status;
   */

   PROCEDURE register_user (email_address_in IN VARCHAR2, id_out OUT NUMBER);

   PROCEDURE register_user (email_address_in IN VARCHAR2);

   FUNCTION register_user (email_address_in IN VARCHAR2)
      RETURN NUMBER;

   /* Specify frequency as:
      # - number of minutes after which to check
      H - Once an hour = #60
      D - Once a day = #60X24
      N - Never

      NOTE: all register_alert procedures perform an upsert -
            if an alert already exists with this name, the
            alert is updated with these characteristics, and
            existing actions are deleted and replaced with
            those provided here.
   */

   FUNCTION is_query_alert_type (type_in IN VARCHAR2)
      RETURN BOOLEAN;

   FUNCTION frequency_description (frequency_in IN VARCHAR2)
      RETURN VARCHAR2;

   PROCEDURE register_alert (
      alert_name_in               IN     VARCHAR2,
      description_in              IN     VARCHAR2 DEFAULT NULL,
      trigger_code_in             IN     VARCHAR2,
      trigger_type_in             IN     VARCHAR2,
      check_frequency_in          IN     VARCHAR2,
      recurrence_frequency_in     IN     VARCHAR2,
      ignore_failures_in          IN     VARCHAR2,
      action_type_in              IN     VARCHAR2 DEFAULT NULL,
      action_code_in              IN     VARCHAR2 DEFAULT NULL,
      send_email_to_in            IN     VARCHAR2 DEFAULT NULL,
      email_subject_in            IN     VARCHAR2 DEFAULT NULL,
      email_body_in               IN     CLOB DEFAULT NULL,
      email_query_in              IN     VARCHAR2 DEFAULT NULL,
      is_active_in                IN     VARCHAR2 DEFAULT 'N',
      owner_name_in               IN     VARCHAR2 DEFAULT NULL,
      table_name_in               IN     VARCHAR2 DEFAULT NULL,
      condition_column1_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator1_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression1_in    IN     VARCHAR2 DEFAULT NULL,
      condition_column2_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator2_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression2_in    IN     VARCHAR2 DEFAULT NULL,
      condition_column3_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator3_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression3_in    IN     VARCHAR2 DEFAULT NULL,
      condition_column4_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator4_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression4_in    IN     VARCHAR2 DEFAULT NULL,
      condition_column5_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator5_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression5_in    IN     VARCHAR2 DEFAULT NULL,
      trigger_key_query_in        IN     VARCHAR2 DEFAULT NULL,
      target_identifier_code_in    IN     VARCHAR2 DEFAULT NULL,
      repeat_for_trigger_key_in   IN     VARCHAR2 DEFAULT NULL,
      repeat_for_target_key_in    IN     VARCHAR2 DEFAULT NULL,
      key_descriptor_code_in      IN     VARCHAR2,
      bind_name_in                IN     VARCHAR2,
      bind_description_in         IN     VARCHAR2,
      bind_datatype_in            IN     VARCHAR2,
      alert_id_out                   OUT NUMBER);

   PROCEDURE register_alert (
      alert_name_in               IN     VARCHAR2,
      description_in              IN     VARCHAR2 DEFAULT NULL,
      trigger_code_in             IN     VARCHAR2,
      trigger_type_in             IN     VARCHAR2,
      check_frequency_in          IN     VARCHAR2,
      recurrence_frequency_in     IN     VARCHAR2,
      ignore_failures_in          IN     VARCHAR2,
      owner_name_in               IN     VARCHAR2 DEFAULT NULL,
      table_name_in               IN     VARCHAR2 DEFAULT NULL,
      condition_column1_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator1_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression1_in    IN     VARCHAR2 DEFAULT NULL,
      condition_column2_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator2_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression2_in    IN     VARCHAR2 DEFAULT NULL,
      condition_column3_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator3_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression3_in    IN     VARCHAR2 DEFAULT NULL,
      condition_column4_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator4_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression4_in    IN     VARCHAR2 DEFAULT NULL,
      condition_column5_in        IN     VARCHAR2 DEFAULT NULL,
      condition_operator5_in      IN     VARCHAR2 DEFAULT NULL,
      condition_expression5_in    IN     VARCHAR2 DEFAULT NULL,
      trigger_key_query_in        IN     VARCHAR2 DEFAULT NULL,
      target_identifier_code_in    IN     VARCHAR2 DEFAULT NULL,
      repeat_for_trigger_key_in   IN     VARCHAR2 DEFAULT NULL,
      repeat_for_target_key_in    IN     VARCHAR2 DEFAULT NULL,
      key_descriptor_code_in      IN     VARCHAR2,
      bind_name_in                IN     VARCHAR2,
      bind_description_in         IN     VARCHAR2,
      bind_datatype_in            IN     VARCHAR2,
      actions_in                  IN     actions_t,
      is_active_in                IN     VARCHAR2 DEFAULT 'N',
      alert_id_out                   OUT NUMBER);

   PROCEDURE register_alert (
      alert_name_in               IN VARCHAR2,
      description_in              IN VARCHAR2 DEFAULT NULL,
      trigger_code_in             IN VARCHAR2,
      trigger_type_in             IN VARCHAR2,
      check_frequency_in          IN VARCHAR2,
      recurrence_frequency_in     IN VARCHAR2,
      ignore_failures_in          IN VARCHAR2,
      action_type_in              IN VARCHAR2 DEFAULT NULL,
      action_code_in              IN VARCHAR2 DEFAULT NULL,
      send_email_to_in            IN VARCHAR2 DEFAULT NULL,
      email_subject_in            IN VARCHAR2 DEFAULT NULL,
      email_body_in               IN CLOB DEFAULT NULL,
      email_query_in              IN VARCHAR2 DEFAULT NULL,
      is_active_in                IN VARCHAR2 DEFAULT 'N',
      owner_name_in               IN VARCHAR2 DEFAULT NULL,
      table_name_in               IN VARCHAR2 DEFAULT NULL,
      condition_column1_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator1_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression1_in    IN VARCHAR2 DEFAULT NULL,
      condition_column2_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator2_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression2_in    IN VARCHAR2 DEFAULT NULL,
      condition_column3_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator3_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression3_in    IN VARCHAR2 DEFAULT NULL,
      condition_column4_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator4_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression4_in    IN VARCHAR2 DEFAULT NULL,
      condition_column5_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator5_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression5_in    IN VARCHAR2 DEFAULT NULL,
      trigger_key_query_in        IN VARCHAR2 DEFAULT NULL,
      target_identifier_code_in    IN VARCHAR2 DEFAULT NULL,
      repeat_for_trigger_key_in   IN VARCHAR2 DEFAULT NULL,
      repeat_for_target_key_in    IN VARCHAR2 DEFAULT NULL,
      key_descriptor_code_in      IN VARCHAR2,
      bind_name_in                IN VARCHAR2,
      bind_description_in         IN VARCHAR2,
      bind_datatype_in            IN VARCHAR2);

   PROCEDURE register_alert (
      alert_name_in               IN VARCHAR2,
      description_in              IN VARCHAR2 DEFAULT NULL,
      trigger_code_in             IN VARCHAR2,
      trigger_type_in             IN VARCHAR2,
      check_frequency_in          IN VARCHAR2,
      recurrence_frequency_in     IN VARCHAR2,
      ignore_failures_in          IN VARCHAR2,
      owner_name_in               IN VARCHAR2 DEFAULT NULL,
      table_name_in               IN VARCHAR2 DEFAULT NULL,
      condition_column1_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator1_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression1_in    IN VARCHAR2 DEFAULT NULL,
      condition_column2_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator2_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression2_in    IN VARCHAR2 DEFAULT NULL,
      condition_column3_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator3_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression3_in    IN VARCHAR2 DEFAULT NULL,
      condition_column4_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator4_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression4_in    IN VARCHAR2 DEFAULT NULL,
      condition_column5_in        IN VARCHAR2 DEFAULT NULL,
      condition_operator5_in      IN VARCHAR2 DEFAULT NULL,
      condition_expression5_in    IN VARCHAR2 DEFAULT NULL,
      trigger_key_query_in        IN VARCHAR2 DEFAULT NULL,
      target_identifier_code_in    IN VARCHAR2 DEFAULT NULL,
      repeat_for_trigger_key_in   IN VARCHAR2 DEFAULT NULL,
      repeat_for_target_key_in    IN VARCHAR2 DEFAULT NULL,
      key_descriptor_code_in      IN VARCHAR2,
      bind_name_in                IN VARCHAR2,
      bind_description_in         IN VARCHAR2,
      bind_datatype_in            IN VARCHAR2,
      actions_in                  IN actions_t,
      is_active_in                IN VARCHAR2 DEFAULT 'N');

   PROCEDURE follow_alert (alert_id_in     IN NUMBER,
                           user_id_in      IN NUMBER,
                           bind_value_in   IN VARCHAR2);

   PROCEDURE follow_alert (alert_id_in        IN NUMBER,
                           email_address_in   IN VARCHAR2,
                           bind_value_in      IN VARCHAR2);

   PROCEDURE follow_alert (alert_name_in      IN VARCHAR2,
                           email_address_in   IN VARCHAR2,
                           bind_value_in      IN VARCHAR2);

   PROCEDURE unfollow_alert (alert_id_in IN NUMBER, user_id_in IN NUMBER);

   PROCEDURE unfollow_alert (alert_name_in      IN VARCHAR2,
                             email_address_in   IN VARCHAR2);

   /* Activate an alert for checking */
   PROCEDURE activate_alert (id_in IN NUMBER);

   /* Ignore this alert when checking */
   PROCEDURE deactivate_alert (id_in IN NUMBER);

   /* Run by job at specified frequency to see if any
    alerts have been triggered */
   PROCEDURE check_alerts;

   FUNCTION constructed_query (alert_id_in IN NUMBER)
      RETURN VARCHAR2;

   FUNCTION time_to_trigger_alert (check_frequency_in          IN VARCHAR2,
                                   recurring_frequency_in      IN VARCHAR2,
                                   completed_at_in             IN DATE,
                                   completed_successfully_in   IN VARCHAR2,
                                   ignore_failures_in          IN VARCHAR2)
      RETURN BOOLEAN;

   PROCEDURE execute_alert_actions (alert_id_in IN NUMBER);

   /* Add to the alert runs tables and also update the alert row with
      the last run id */
   PROCEDURE record_result (
      alert_id_in                 IN NUMBER,
      completed_at_in             IN DATE DEFAULT SYSDATE,
      completed_successfully_in   IN VARCHAR2 DEFAULT 'Y',
      results_in                  IN VARCHAR2 DEFAULT NULL,
      errors_in                   IN VARCHAR2 DEFAULT NULL);

   FUNCTION one_alert (id_in IN NUMBER)
      RETURN alerts%ROWTYPE;

   FUNCTION one_alert_follow (alert_id_in IN NUMBER, user_id_in in number)
      RETURN alert_follows%ROWTYPE;

   FUNCTION last_run_for (alert_id_in IN NUMBER)
      RETURN alert_runs%ROWTYPE;

   /* Advanced (BYKEY) alert functionality */
   PROCEDURE record_action_taken (
      alert_run_id_in        IN NUMBER,
      trigger_key_value_in   IN VARCHAR2,
      target_key_value_in    IN VARCHAR2 DEFAULT NULL);

   FUNCTION action_taken (alert_id_in            IN NUMBER,
                          trigger_key_value_in   IN VARCHAR2,
                          target_key_value_in    IN VARCHAR2 DEFAULT NULL)
      RETURN BOOLEAN;

   FUNCTION action_taken_yn (alert_id_in            IN NUMBER,
                             trigger_key_value_in   IN VARCHAR2,
                             target_key_value_in    IN VARCHAR2 DEFAULT NULL)
      RETURN VARCHAR2;

   FUNCTION is_simple_alert (alert_id_in IN NUMBER)
      RETURN BOOLEAN;

   FUNCTION is_simple_alert_yn (alert_id_in IN NUMBER)
      RETURN VARCHAR2;
END;
/