/* Table used for demonstrations */

CREATE TABLE bc_tips
(
   tip_id     INTEGER,
   tip_name   VARCHAR2 (100)
)
/

INSERT INTO bc_tips
       SELECT LEVEL, 'Here''s another great tip!'
         FROM DUAL
   CONNECT BY LEVEL <= 1000
/

COMMIT
/

/* Can use all three types of collections. */

CREATE OR REPLACE PACKAGE bc_colls
   AUTHID DEFINER
IS
   TYPE associative_array_t IS TABLE OF bc_tips%ROWTYPE
      INDEX BY PLS_INTEGER;

   TYPE nested_table_t IS TABLE OF bc_tips%ROWTYPE;

   TYPE array_t IS VARRAY (1000) OF bc_tips%ROWTYPE;
END;
/

DECLARE
   l_aa   bc_colls.associative_array_t;
BEGIN
   SELECT *
     BULK COLLECT INTO l_aa
     from bc_tips;

   DBMS_OUTPUT.put_line (l_aa.COUNT);
END;
/

DECLARE
   l_nt   bc_colls.nested_table_t;
BEGIN
   SELECT *
     BULK COLLECT INTO l_nt
     from bc_tips;

   DBMS_OUTPUT.put_line (l_nt.COUNT);
END;
/

DECLARE
   l_va   bc_colls.array_t;
BEGIN
   SELECT *
     BULK COLLECT INTO l_va
     from bc_tips;

   DBMS_OUTPUT.put_line (l_va.COUNT);
END;
/

/* No need to initialize and extend. */

DECLARE
   l_nt   bc_colls.nested_table_t;
BEGIN
   l_nt (1).tip_id := 1;
END;
/

DECLARE
   l_nt   bc_colls.nested_table_t := bc_colls.nested_table_t ();
BEGIN
   l_nt (1).tip_id := 1;
END;
/

DECLARE
   l_nt   bc_colls.nested_table_t := bc_colls.nested_table_t ();
BEGIN
   l_nt.EXTEND;
   l_nt (1).tip_id := 1;
END;
/

DECLARE
   l_nt   bc_colls.nested_table_t;
BEGIN
   SELECT *
     BULK COLLECT INTO l_nt
     from bc_tips;

   DBMS_OUTPUT.put_line (l_nt.COUNT);
END;
/

/* Can bulk collect into collection of scalars or records. */

DECLARE
   TYPE my_tip_ids_t IS TABLE OF bc_tips.tip_id%TYPE
      INDEX BY PLS_INTEGER;

   l_tip_ids     my_tip_ids_t;

   TYPE my_tip_names_t IS TABLE OF bc_tips.tip_name%TYPE
      INDEX BY PLS_INTEGER;

   l_tip_names   my_tip_names_t;

   TYPE my_associative_array_t IS TABLE OF bc_tips%ROWTYPE
      INDEX BY PLS_INTEGER;

   l_aa          my_associative_array_t;
BEGIN
   SELECT tip_id, tip_name
     BULK COLLECT INTO l_tip_ids, l_tip_names
     from bc_tips;

   DBMS_OUTPUT.put_line (l_tip_ids.COUNT);

   SELECT *
     BULK COLLECT INTO l_aa
     from bc_tips;

   DBMS_OUTPUT.put_line (l_aa.COUNT);
END;
/

/* collection contents alwayrs replaced with each BULK COLLECT. */

DECLARE
   l_aa   bc_colls.associative_array_t;
BEGIN
     SELECT *
       BULK COLLECT INTO l_aa
       from bc_tips
   ORDER BY tip_id ASC;

   DBMS_OUTPUT.put_line (l_aa.COUNT || '-' || l_aa (l_aa.FIRST).tip_id);

     SELECT *
       BULK COLLECT INTO l_aa
       from bc_tips
   ORDER BY tip_id DESC;

   DBMS_OUTPUT.put_line (l_aa.COUNT || '-' || l_aa (l_aa.FIRST).tip_id);
END;
/


DECLARE
   l_aa   bc_colls.associative_array_t;

   CURSOR tips_cur
   IS
        SELECT *
          from bc_tips
      ORDER BY tip_id ASC;
BEGIN
   OPEN tips_cur;

   LOOP
      FETCH tips_cur BULK COLLECT INTO l_aa LIMIT 100;

      EXIT WHEN l_aa.COUNT = 0;

      DBMS_OUTPUT.put_line (l_aa.COUNT || '-' || l_aa (l_aa.FIRST).tip_id);
   END LOOP;
END;
/

/* Collection always sequentially filled. */

DECLARE
   l_aa   bc_colls.associative_array_t;
BEGIN
   SELECT *
     BULK COLLECT INTO l_aa
     from bc_tips;

   DBMS_OUTPUT.put_line ('Count = ' || l_aa.COUNT);

   DBMS_OUTPUT.put_line ('First = ' || l_aa.FIRST);

   DBMS_OUTPUT.put_line ('Last = ' || l_aa.LAST);
   
   /* Always true */
   IF l_aa.COUNT = NVL ((l_aa.LAST - l_aa.FIRST), 0) + 1
   THEN
      DBMS_OUTPUT.put_line ('Always True!');
   END IF;
END;
/

DROP TABLE bc_tips
/
