/* Assumption: you have already created the EMPLOYEES table */

/* Now create an error log table for failed DML operations */

BEGIN
   DBMS_ERRLOG.create_error_log (dml_table_name => 'EMPLOYEES');
END;
/

/* There are five error-related columns (starting with ORA_ERR*)
   and then a column for each VARCHAR2-compatible column in the
   DML table.
*/

  SELECT column_name, data_type, data_length
    FROM user_tab_columns
   WHERE table_name = 'ERR$_EMPLOYEES'
ORDER BY column_id
/

/*
Now let's try to do an update with some bad data.
This will fail because multiplying salary by 200 will
generate some values that are too big to fit into the 
column. 

First, without LOG ERRORS....which means the DML is 
"all or nothing" - no rows are updated if any one row
fails to update.
*/

DECLARE
   l_count   PLS_INTEGER;
BEGIN
   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE salary > 24000;

   DBMS_OUTPUT.put_line ('Before ' || l_count);

   UPDATE employees
      SET salary = salary * 200;

   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE salary > 24000;
EXCEPTION
   WHEN OTHERS
   THEN
      DBMS_OUTPUT.put_line (DBMS_UTILITY.format_error_stack);

      SELECT COUNT (*)
        INTO l_count
        FROM employees
       WHERE salary > 24000;

      DBMS_OUTPUT.put_line ('After ' || l_count);
END;
/

/* And now with LOG ERRORS: specify unlimited # of errors, OK. */

DECLARE
   l_count   PLS_INTEGER;
BEGIN
   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE salary > 24000;

   DBMS_OUTPUT.put_line ('Before ' || l_count);

       UPDATE employees
          SET salary = salary * 200
   LOG ERRORS REJECT LIMIT UNLIMITED;

   DBMS_OUTPUT.put_line ('After 1 ' || SQL%ROWCOUNT);

   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE salary > 24000;

   DBMS_OUTPUT.put_line ('After 2 ' || l_count);

   ROLLBACK;
END;
/

/*
When using LOG ERRORS, you *must* check the error log table
immediately after running your DML statement. No error will
be raised to indicate there was any problem.
*/

SELECT 'Number of errors = ' || COUNT (*) FROM err$_employees
/

SELECT ora_err_mesg$, last_name FROM err$_employees
/

BEGIN
   DELETE FROM err$_employees;

   COMMIT;
END;
/

/* Try to triple the size of the first name
   and allow no more than 10 errors. */

BEGIN
       UPDATE employees
          SET first_name = first_name || first_name || first_name
   LOG ERRORS REJECT LIMIT 10;

   ROLLBACK;
END;
/

SELECT 'Number of errors = ' || COUNT (*) FROM err$_employees
/