CREATE OR REPLACE PACKAGE BODY dg_workout_mgr
/*
===========================
Oracle Dev Gym
===========================
Copyright 2016 Oracle Corporation
Manage goals and workouts
*/
IS
   FUNCTION is_dg_resource (resource_usage_in IN VARCHAR2)
      RETURN BOOLEAN
   IS
   BEGIN
      RETURN resource_usage_in IN ('Quiz', 'Challenge', 'Test');
   END;

   FUNCTION end_from_start_date (start_date_in IN DATE)
      RETURN DATE
   IS
   BEGIN
      RETURN qdb_utilities.end_of_day (
                NEXT_DAY (
                   NEXT_DAY (
                      NEXT_DAY (NEXT_DAY (start_date_in, 'FRI'), 'FRI'),
                      'FRI'),
                   'FRI'));
   END;

   FUNCTION first_workout_start_date (
      immediate_start_in   IN BOOLEAN DEFAULT TRUE)
      RETURN DATE
   IS
   BEGIN
      /* If Sat - Tue, then started on that (previous) Sat.
      Otherwise, next Sat. Unless immediate start desired. */

      RETURN qdb_utilities.start_of_day (
                NEXT_DAY (
                     SYSDATE
                   - CASE
                        WHEN    immediate_start_in
                             OR TO_CHAR (SYSDATE, 'D') <= 3
                        THEN
                           5
                        ELSE
                           0
                     END,
                   'SAT'));
   END;

   FUNCTION last_workout_end_date (
      immediate_start_in   IN BOOLEAN DEFAULT TRUE)
      RETURN DATE
   IS
   BEGIN
      /* Fourth Friday after first start date. */

      RETURN end_from_start_date (
                start_date_in   => first_workout_start_date (
                                     immediate_start_in));
   END;

   FUNCTION playing_daily_db_quiz (user_id_in IN INTEGER)
      RETURN BOOLEAN
   IS
   BEGIN
      RETURN TRUE;
   END;

   FUNCTION one_goal (goal_id_in IN INTEGER)
      RETURN dg_goals%ROWTYPE
   IS
      l_goal   dg_goals%ROWTYPE;
   BEGIN
      SELECT *
        INTO l_goal
        FROM dg_goals g
       WHERE g.goal_id = goal_id_in;

      RETURN l_goal;
   EXCEPTION
      WHEN NO_DATA_FOUND
      THEN
         RETURN NULL;
   END;

   FUNCTION one_workout (workout_id_in IN INTEGER)
      RETURN dg_workouts%ROWTYPE
   IS
      l_workout   dg_workouts%ROWTYPE;
   BEGIN
      SELECT *
        INTO l_workout
        FROM dg_workouts g
       WHERE g.workout_id = workout_id_in;

      RETURN l_workout;
   EXCEPTION
      WHEN NO_DATA_FOUND
      THEN
         RETURN NULL;
   END;

   FUNCTION current_goal (user_id_in IN INTEGER)
      RETURN dg_goals%ROWTYPE
   IS
      l_goal   dg_goals%ROWTYPE;

      CURSOR nearest_goal_cur (
         date_in   IN DATE)
      IS
         SELECT *
           FROM dg_goals g
          WHERE     g.user_id = user_id_in
                AND date_in BETWEEN g.start_date AND g.End_date;
   BEGIN
      OPEN nearest_goal_cur (SYSDATE);

      FETCH nearest_goal_cur INTO l_goal;

      CLOSE nearest_goal_cur;

      IF l_goal.goal_id IS NULL
      THEN
         OPEN nearest_goal_cur (SYSDATE + 7);

         FETCH nearest_goal_cur INTO l_goal;

         CLOSE nearest_goal_cur;
      END IF;

      RETURN l_goal;
   END;

   FUNCTION last_completed_goal (user_id_in IN INTEGER)
      RETURN dg_goals%ROWTYPE
   IS
      l_goal   dg_goals%ROWTYPE;
   BEGIN
      SELECT *
        INTO l_goal
        FROM dg_goals g
       WHERE     g.user_id = user_id_in
             AND g.end_date = (SELECT MAX (end_date)
                                 FROM dg_goals
                                WHERE SYSDATE > g.End_date);

      RETURN l_goal;
   EXCEPTION
      WHEN NO_DATA_FOUND
      THEN
         RETURN NULL;
   END;

   FUNCTION current_goal_id (user_id_in IN INTEGER)
      RETURN INTEGER
   IS
      l_goal   dg_goals%ROWTYPE := current_goal (user_id_in);
   BEGIN
      RETURN l_goal.goal_id;
   END;

   FUNCTION current_goal_start_date (user_id_in IN INTEGER)
      RETURN DATE
   IS
      l_goal   dg_goals%ROWTYPE := current_goal (user_id_in);
   BEGIN
      RETURN l_goal.start_date;
   END;

   FUNCTION current_goal_end_date (user_id_in IN INTEGER)
      RETURN DATE
   IS
      l_goal   dg_goals%ROWTYPE := current_goal (user_id_in);
   BEGIN
      RETURN l_goal.end_date;
   END;

   FUNCTION current_goal_completion_pct (user_id_in IN INTEGER)
      RETURN INTEGER
   IS
      l_goal         dg_goals%ROWTYPE := current_goal (user_id_in);
      l_quiz_count   INTEGER;
   BEGIN
      SELECT COUNT (*)
        INTO l_quiz_count
        FROM qdb_compev_answers
       WHERE     user_id = user_id_in
             AND ended_on BETWEEN l_goal.start_date AND l_goal.end_date;

      RETURN ROUND ( (l_quiz_count / l_goal.target_min_quizzes) * 100);
   END;

   FUNCTION current_wo_completion_pct (user_id_in IN INTEGER)
      RETURN INTEGER
   IS
      l_workout_id   dg_workouts.workout_id%TYPE
                        := current_workout_id (user_id_in);
      l_completed    INTEGER;
      l_total        INTEGER;
   BEGIN
      SELECT COUNT (*),
             SUM (CASE WHEN completed_date IS NOT NULL THEN 1 ELSE 0 END)
        INTO l_total, l_completed
        FROM dg_workout_activities
       WHERE workout_id = l_workout_id;

      RETURN ROUND ( (l_completed / l_total) * 100);
   END;

   PROCEDURE get_next_goal_dates (user_id_in       IN     INTEGER,
                                  start_date_out      OUT DATE,
                                  end_date_out        OUT DATE)
   IS
   BEGIN
      start_date_out := current_goal_end_date (user_id_in) + 1;
      end_date_out := end_from_start_date (start_date_out);
   END;

   PROCEDURE get_goal_dates (goal_id_in       IN     INTEGER,
                             start_date_out      OUT DATE,
                             end_date_out        OUT DATE)
   IS
   BEGIN
      SELECT start_date, end_date
        INTO start_date_out, end_date_out
        FROM dg_goals
       WHERE goal_id = goal_id_in;
   END;

   FUNCTION next_goal_id (user_id_in IN INTEGER)
      RETURN INTEGER
   IS
      c_date   CONSTANT DATE := current_goal_end_date (user_id_in);

      CURSOR user_goals_cur
      IS
           SELECT goal_id
             FROM dg_goals
            WHERE user_id = user_id_in AND start_date > c_date
         ORDER BY start_date;

      l_goal_id         dg_goals.goal_id%TYPE;
   BEGIN
      OPEN user_goals_cur;

      FETCH user_goals_cur INTO l_goal_id;

      CLOSE user_goals_cur;

      RETURN l_goal_id;
   END;

   PROCEDURE insert_goal (user_id_in                IN     INTEGER,
                          target_pct_correct_in     IN     INTEGER,
                          target_min_quizzes_in     IN     INTEGER,
                          target_daily_ranking_in   IN     INTEGER,
                          intensity_level_in        IN     VARCHAR2,
                          play_daily_db_quiz_in     IN     VARCHAR2,
                          start_date_in             IN     DATE DEFAULT NULL,
                          end_date_in               IN     DATE DEFAULT NULL,
                          goal_id_out                  OUT INTEGER)
   /*
   If dates not provided, then create this for the next available period.
   */
   IS
      l_start   DATE := NVL (start_date_in, first_workout_start_date (TRUE));
      l_end     DATE := NVL (end_date_in, last_workout_end_date (TRUE));
   BEGIN
      INSERT INTO dg_goals (user_id,
                            target_pct_correct,
                            target_daily_ranking,
                            target_min_quizzes,
                            intensity_level,
                            play_daily_db_quiz,
                            start_date,
                            end_date,
                            goal_period)
           VALUES (user_id_in,
                   target_pct_correct_in,
                   target_daily_ranking_in,
                   target_min_quizzes_in,
                   intensity_level_in,
                   play_daily_db_quiz_in,
                   l_start,
                   l_end,
                   'M')
        RETURNING goal_id
             INTO goal_id_out;
   END;

   PROCEDURE update_goal (goal_id_in                IN INTEGER,
                          target_pct_correct_in     IN INTEGER,
                          target_min_quizzes_in     IN INTEGER,
                          target_daily_ranking_in   IN INTEGER,
                          intensity_level_in        IN VARCHAR2,
                          play_daily_db_quiz_in     IN VARCHAR2,
                          start_date_in             IN DATE DEFAULT NULL,
                          end_date_in               IN DATE DEFAULT NULL,
                          topic1_id_in              IN INTEGER DEFAULT NULL,
                          topic2_id_in              IN INTEGER DEFAULT NULL,
                          topic3_id_in              IN INTEGER DEFAULT NULL,
                          topic4_id_in              IN INTEGER DEFAULT NULL,
                          topic1_difficulty_id_in   IN INTEGER DEFAULT NULL,
                          topic2_difficulty_id_in   IN INTEGER DEFAULT NULL,
                          topic3_difficulty_id_in   IN INTEGER DEFAULT NULL,
                          topic4_difficulty_id_in   IN INTEGER DEFAULT NULL)
   IS
      l_start     DATE;
      l_user_id   INTEGER;
   BEGIN
         UPDATE dg_goals
            SET target_pct_correct = target_pct_correct_in,
                target_daily_ranking = target_daily_ranking_in,
                target_min_quizzes = target_min_quizzes_in,
                intensity_level = intensity_level_in,
                play_daily_db_quiz = play_daily_db_quiz_in
          WHERE goal_id = goal_id_in
      RETURNING start_date, user_id
           INTO l_start, l_user_id;

      set_workout (goal_id_in           => goal_id_in,
                   user_id_in           => l_user_id,
                   topic_id_in          => topic1_id_in,
                   week_number_in       => 1,
                   intensity_level_in   => intensity_level_in,
                   difficulty_id_in     => topic1_difficulty_id_in,
                   goal_start_date_in   => l_start);

      set_workout (goal_id_in           => goal_id_in,
                   user_id_in           => l_user_id,
                   topic_id_in          => topic2_id_in,
                   week_number_in       => 2,
                   intensity_level_in   => intensity_level_in,
                   difficulty_id_in     => topic2_difficulty_id_in,
                   goal_start_date_in   => l_start);

      set_workout (goal_id_in           => goal_id_in,
                   user_id_in           => l_user_id,
                   topic_id_in          => topic3_id_in,
                   week_number_in       => 3,
                   intensity_level_in   => intensity_level_in,
                   difficulty_id_in     => topic3_difficulty_id_in,
                   goal_start_date_in   => l_start);

      set_workout (goal_id_in           => goal_id_in,
                   user_id_in           => l_user_id,
                   topic_id_in          => topic4_id_in,
                   week_number_in       => 4,
                   intensity_level_in   => intensity_level_in,
                   difficulty_id_in     => topic4_difficulty_id_in,
                   goal_start_date_in   => l_start);
   END;

   PROCEDURE get_goal (goal_id_in                 IN     INTEGER,
                       target_pct_correct_out        OUT INTEGER,
                       target_min_quizzes_out        OUT INTEGER,
                       target_daily_ranking_out      OUT INTEGER,
                       intensity_level_out           OUT VARCHAR2,
                       play_daily_db_quiz_out        OUT VARCHAR2,
                       start_date_out                OUT DATE,
                       end_date_out                  OUT DATE,
                       topic1_id_out                 OUT INTEGER,
                       topic2_id_out                 OUT INTEGER,
                       topic3_id_out                 OUT INTEGER,
                       topic4_id_out                 OUT INTEGER,
                       topic1_difficulty_id_out      OUT INTEGER,
                       topic2_difficulty_id_out      OUT INTEGER,
                       topic3_difficulty_id_out      OUT INTEGER,
                       topic4_difficulty_id_out      OUT INTEGER)
   IS
      l_topics         DBMS_SQL.number_table;
      l_difficulties   DBMS_SQL.number_table;
   BEGIN
      SELECT target_pct_correct,
             target_daily_ranking,
             target_min_quizzes,
             intensity_level,
             play_daily_db_quiz,
             start_date,
             end_date
        INTO target_pct_correct_out,
             target_daily_ranking_out,
             target_min_quizzes_out,
             intensity_level_out,
             play_daily_db_quiz_out,
             start_date_out,
             end_date_out
        FROM dg_goals
       WHERE goal_id = goal_id_in;

        SELECT topic_id, difficulty_id
          BULK COLLECT INTO l_topics, l_difficulties
          FROM dg_workouts
         WHERE goal_id = goal_id_in
      ORDER BY week_number;

      IF l_topics.COUNT > 0
      THEN
         /* There are always either none or 4 */
         topic1_id_out := l_topics (1);
         topic1_difficulty_id_out := l_difficulties (1);
         topic2_id_out := l_topics (2);
         topic2_difficulty_id_out := l_difficulties (2);
         topic3_id_out := l_topics (3);
         topic3_difficulty_id_out := l_difficulties (3);
         topic4_id_out := l_topics (4);
         topic4_difficulty_id_out := l_difficulties (4);
      END IF;
   END;

   PROCEDURE get_goal_actuals (goal_id_in          IN     INTEGER,
                               pct_correct_out        OUT NUMBER,
                               quiz_counts_out        OUT INTEGER,
                               daily_ranking_out      OUT INTEGER)
   IS
      c_goal   CONSTANT dg_goals%ROWTYPE := one_goal (goal_id_in);
      l_d3_taken        INTEGER;
   BEGIN
      SELECT ROUND (AVG (pct_correct_answers), 1),
             COUNT (eva.compev_answer_id)
        INTO pct_correct_out, quiz_counts_out
        FROM qdb_compev_answers_v eva
       WHERE eva.ended_on BETWEEN c_goal.start_date AND c_goal.end_Date;

      SELECT COUNT (*)
        INTO l_d3_taken
        FROM mv_qdb_rankings_d, dg_goals g
       WHERE     competition_id = dg_daily_quiz_mgr.daily_quiz_comp_id
             AND date_used BETWEEN c_goal.start_date AND c_goal.end_Date
             AND overall_rank > c_goal.target_daily_ranking;
   END;

   PROCEDURE turn_on_db_quiz_play (user_id_in IN INTEGER)
   IS
   BEGIN
      UPDATE dg_goals
         SET play_daily_db_quiz = qdb_config.c_yes
       WHERE user_id = user_id_in AND SYSDATE BETWEEN start_date AND end_date;
   END;

   PROCEDURE copy_last_goal (user_id_in IN INTEGER, goal_id_out OUT INTEGER)
   IS
      l_new_goal   dg_goals%ROWTYPE := last_completed_goal (user_id_in);
   BEGIN
      /* Create a new goal for next period from previous one */
      l_new_goal.goal_id := NULL;
      l_new_goal.start_date := first_workout_start_date ();
      l_new_goal.end_date := last_workout_end_date ();

      INSERT INTO dg_goals
           VALUES l_new_goal
        RETURNING goal_id
             INTO goal_id_out;
   END;

   PROCEDURE set_domain_goal (goal_id_in              IN INTEGER,
                              domain_id_in               INTEGER,
                              target_min_ranking_in   IN INTEGER,
                              target_pct_correct_in   IN INTEGER,
                              min_difficulty_id_in    IN INTEGER)
   IS
   /* If one already exists, replace it */
   BEGIN
      INSERT INTO dg_domain_goals (goal_id,
                                   domain_id,
                                   target_min_ranking,
                                   target_pct_correct,
                                   min_difficulty_id)
           VALUES (goal_id_in,
                   domain_id_in,
                   target_min_ranking_in,
                   target_pct_correct_in,
                   min_difficulty_id_in);
   EXCEPTION
      WHEN DUP_VAL_ON_INDEX
      THEN
         UPDATE dg_domain_goals
            SET target_min_ranking = target_min_ranking_in,
                target_pct_correct = target_pct_correct,
                min_difficulty_id = min_difficulty_id_in
          WHERE goal_id = goal_id_in AND domain_id = domain_id_in;
   END;

   FUNCTION minutes_for_activity_type (duration_mins_in      IN NUMBER,
                                       activity_type_id_in   IN INTEGER)
      RETURN NUMBER
   IS
      CURSOR info_cur
      IS
         SELECT minimum_minutes, padding_formula
           FROM dg_activity_typeS
          WHERE activity_type_id = activity_type_id_in;

      info_rec   info_cur%ROWTYPE;
      l_return   NUMBER;
   BEGIN
      OPEN info_cur;

      FETCH info_cur INTO info_rec;

      CLOSE info_cur;

      EXECUTE IMMEDIATE
            'begin :mins := LEAST (:resmin, :atmin) '
         || info_rec.padding_formula
         || '; end;'
         USING OUT l_return, duration_mins_in, info_rec.minimum_minutes;

      /* Absolute minimum: 5 minutes */

      RETURN NVL (l_return, 5);
   END;

   FUNCTION activities_for_topic (topic_id_in          IN INTEGER,
                                  intensity_level_in   IN VARCHAR2,
                                  difficulty_id_in     IN INTEGER)
      RETURN dg_workout_activities_nt
   IS
      c_target_time         CONSTANT NUMBER
                                        := minutes_for_level (intensity_level_in) ;
      c_difficulty_number   CONSTANT NUMBER
         := qdb_content.difficulty_number (difficulty_id_in) ;
      l_activity_id                  INTEGER;
      l_quiz_type                    INTEGER;
      l_counter                      INTEGER := 1;
      l_total_time                   NUMBER := 0;
      l_activities                   dg_workout_activities_nt
                                        := dg_workout_activities_nt ();

      /* All resources for that topic and everything below it.
      First priority: resources with "after" resources - likely quizzes.
      Second priority: videos then tutorials the quizzes
      */

      CURSOR resources_cur
      IS
           SELECT r.*
             FROM ov.ov_resources_external r, qdb_topics t
            WHERE     r.topic_id = t.ch_topic_id
                  AND t.topic_id IN (    SELECT tr.topic_id
                                           FROM qdb_topics tr
                                     CONNECT BY PRIOR tr.topic_id =
                                                   tr.parent_topic_id
                                     START WITH tr.topic_id = topic_id_in)
                  AND (   r.difficulty_number = c_difficulty_number
                       OR r.difficulty_number IS NULL)
         ORDER BY CASE WHEN after_count > 0 THEN 0 ELSE 1 END,
                  CASE
                     WHEN r.resource_usage = 'Video'
                     THEN
                        1
                     WHEN r.resource_usage IN ('OBE', 'Tutorial')
                     THEN
                        2
                     /* Could call is_dg_resource (r.resource_usage) but context switches! */
                     WHEN r.resource_usage IN ('Quiz', 'Challenge', 'Test')
                     THEN
                        3
                     ELSE
                        4
                  END;

      resources_rec                  ov.ov_resources_external%ROWTYPE;

      PROCEDURE add_one_resource (
         resource_in            IN     ov.ov_resources_external%ROWTYPE,
         after_activity_id_in   IN     INTEGER,
         total_time_io          IN OUT NUMBER,
         counter_io             IN OUT NUMBER)
      IS
         c_is_dg_resource   CONSTANT PLS_INTEGER
            := CASE
                  WHEN is_dg_resource (resource_in.resource_usage) THEN 1
                  ELSE 0
               END ;
         l_duration                  NUMBER;
      BEGIN
         IF qdb_utilities.trace_is_on
         THEN
            qdb_utilities.trace_activity (
               'add_one_resource isDG-URL-key',
                  c_is_dg_resource
               || '-'
               || CASE
                     WHEN c_is_dg_resource = 1
                     THEN
                        qdb_utilities.apex_url (
                           app_id_in             => 'ORA_DEV_GYM',
                           app_pg_id_in          => 5,
                           app_session_in        => NULL,
                           request_in            => NULL,
                           debug_in              => 'NO',
                           cache_in              => 5,
                           page_item_in          => 'P5_QUESTION_ID',
                           value_in              => resource_in.external_resource_key,
                           add_href_in           => FALSE,
                           return_as_button_in   => FALSE,
                           href_name_in          => NULL,
                           external_url_in       => FALSE,
                           open_new_window_in    => FALSE,
                           onclick_in            => NULL,
                           add_ssp_in            => TRUE)
                     ELSE
                        resource_in.url
                  END
               || '-'
               || resource_in.external_resource_key);
         END IF;

         /* If a DG quiz, test, challenge, then use external_resource_key, generate URL */
         l_duration :=
            minutes_for_activity_type (
               duration_mins_in      => resources_rec.duration_mins,
               activity_type_id_in   => NULL);

         l_activities.EXTEND;

         l_activities (l_activities.LAST) :=
            dg_workout_activities_ot (
               NULL,
               NULL,
               act_type_from_res_usage (resource_in.resource_usage),
               resource_in.resource_id,
               NVL (resource_in.external_resource_key, -1),
               l_counter,
                  CASE WHEN after_activity_id_in IS NOT NULL THEN '^ ' END
               || resource_in.resource_name,
               CASE
                  WHEN     c_is_dg_resource = 1
                       AND resource_in.external_resource_key IS NOT NULL
                  THEN
                     qdb_utilities.apex_url (
                        app_id_in             => 'ORA_DEV_GYM',
                        app_pg_id_in          => 5,
                        app_session_in        => NULL,
                        request_in            => NULL,
                        debug_in              => 'NO',
                        cache_in              => 5,
                        page_item_in          => 'P5_QUESTION_ID',
                        value_in              => resource_in.external_resource_key,
                        add_href_in           => FALSE,
                        return_as_button_in   => FALSE,
                        href_name_in          => NULL,
                        external_url_in       => FALSE,
                        open_new_window_in    => FALSE,
                        onclick_in            => NULL,
                        add_ssp_in            => TRUE)
                  ELSE
                     resource_in.url
               END,
               resource_in.source_name,
               resource_in.source_url,
               after_activity_id_in,
               CASE resource_in.difficulty_number
                  WHEN 1 THEN 'Beginner'
                  WHEN 2 THEN 'Intermediate'
                  WHEN 3 THEN 'Advanced'
               END,
               l_duration,
               NULL,
               NULL,
               NULL,
               'GOAL');
         total_time_io := total_time_io + l_duration;
         counter_io := counter_io + 1;
      END;

      PROCEDURE add_after_resources (parent_resource_id_in   IN     INTEGER,
                                     after_activity_id_in    IN     INTEGER,
                                     total_time_io           IN OUT NUMBER,
                                     counter_io              IN OUT NUMBER)
      IS
         /* Max two quizzes per resource for now. Spread 'em out! */
         l_count     INTEGER := 0;

         CURSOR after_cur
         IS
              SELECT r.*
                FROM ov.ov_resources_external r, ov.ov_resources_related rr
               WHERE     r.resource_id = rr.related_resource_id
                     AND rr.primary_resource_id = parent_resource_id_in
            ORDER BY r.resource_sequence;

         after_rec   ov.ov_resources_external%ROWTYPE;
      BEGIN
         OPEN after_cur;

         LOOP
            FETCH after_cur INTO after_rec;

            EXIT WHEN after_cur%NOTFOUND;

            /* Don't worry about time limit on "after". They are part and parcel,
               and anyway just doing two per.
            EXIT WHEN total_time_io > c_target_time OR after_cur%NOTFOUND; */

            IF NOT resource_in_workout (activities_in   => l_activities,
                                        resource_in     => after_rec)
            THEN
               add_one_resource (
                  resource_in            => after_rec,
                  after_activity_id_in   => after_activity_id_in,
                  total_time_io          => total_time_io,
                  counter_io             => counter_io);
               l_count := l_count + 1;
               EXIT WHEN l_count = 2;
            END IF;
         END LOOP;

         CLOSE after_cur;
      END;

      PROCEDURE add_related_quizzes (topic_id_in            IN     INTEGER,
                                     after_activity_id_in   IN     INTEGER,
                                     total_time_io          IN OUT NUMBER,
                                     counter_io             IN OUT NUMBER)
      IS
         /* Max two quizzes per resource for now. Spread 'em out! */
         l_count         INTEGER := 0;
         l_ch_topic_id   INTEGER;
         l_topic_ids     numbers_nt;
         related_cur     SYS_REFCURSOR;

         related_rec     ov.ov_resources_external%ROWTYPE;
      BEGIN
         SELECT ch_topic_id
           INTO l_ch_topic_id
           FROM qdb_topics
          WHERE topic_id = topic_id_in;

             SELECT tt.topic_id
               BULK COLLECT INTO l_topic_ids
               FROM ov.ov_topics tt
         CONNECT BY PRIOR tt.topic_id = tt.parent_topic_id
         START WITH tt.topic_id = l_ch_topic_id;

         OPEN related_cur FOR
              SELECT r.*
                FROM ov.ov_resources_external r
               WHERE     r.topic_id IN (SELECT COLUMN_VALUE
                                          FROM TABLE (l_topic_ids))
                     AND r.resource_usage = 'Quiz'
            ORDER BY DBMS_RANDOM.VALUE ();

         LOOP
            FETCH related_cur INTO related_rec;

            EXIT WHEN related_cur%NOTFOUND;

            /* Don't worry about time limit on "after". They are part and parcel,
               and anyway just doing two per.
            EXIT WHEN total_time_io > c_target_time OR after_cur%NOTFOUND; */

            IF NOT resource_in_workout (activities_in   => l_activities,
                                        resource_in     => related_rec)
            THEN
               add_one_resource (
                  resource_in            => related_rec,
                  after_activity_id_in   => after_activity_id_in,
                  total_time_io          => total_time_io,
                  counter_io             => counter_io);

               l_count := l_count + 1;
               EXIT WHEN l_count = 2;
            END IF;
         END LOOP;

         CLOSE related_cur;
      END;
   BEGIN
      OPEN resources_cur;

      LOOP
         FETCH resources_cur INTO resources_rec;

         EXIT WHEN l_total_time > c_target_time OR resources_cur%NOTFOUND;

         IF NOT resource_in_workout (activities_in   => l_activities,
                                     resource_in     => resources_rec)
         THEN
            add_one_resource (resource_in            => resources_rec,
                              after_activity_id_in   => NULL,
                              total_time_io          => l_total_time,
                              counter_io             => l_counter);

            IF resources_rec.after_count > 0
            THEN
               add_after_resources (
                  parent_resource_id_in   => resources_rec.resource_id,
                  after_activity_id_in    => l_activity_id,
                  total_time_io           => l_total_time,
                  counter_io              => l_counter);
            ELSE
               /* Add up to two quizzes for the same topic (and below) */
               add_related_quizzes (topic_id_in            => topic_id_in,
                                    after_activity_id_in   => l_activity_id,
                                    total_time_io          => l_total_time,
                                    counter_io             => l_counter);
            END IF;
         END IF;
      END LOOP;

      CLOSE resources_cur;

      RETURN l_activities;
   END;

   PROCEDURE add_workout_activities (workout_id_in        IN INTEGER,
                                     topic_id_in          IN INTEGER,
                                     intensity_level_in   IN VARCHAR2,
                                     difficulty_id_in     IN INTEGER)
   IS
      l_activities   dg_workout_activities_nt := dg_workout_activities_nt ();
   BEGIN
      l_activities :=
         activities_for_topic (topic_id_in,
                               intensity_level_in,
                               difficulty_id_in);

      /* Clear out current entries */
      DELETE FROM dg_workout_activities
            WHERE workout_id = workout_id_in;

      FORALL indx IN 1 .. l_activities.COUNT
         INSERT INTO dg_workout_activities (workout_id,
                                            activity_type_id,
                                            activity_name,
                                            url,
                                            activity_key_id,
                                            activity_sequence,
                                            activity_source,
                                            activity_source_url,
                                            creation_type,
                                            after_activity_id,
                                            estimated_minutes,
                                            difficulty_level)
              VALUES (workout_id_in,
                      l_activities (indx).activity_type_id,
                      l_activities (indx).activity_name,
                      l_activities (indx).url,
                      l_activities (indx).activity_key_id,
                      l_activities (indx).activity_sequence,
                      l_activities (indx).activity_source,
                      l_activities (indx).activity_source_url,
                      l_activities (indx).creation_type,
                      l_activities (indx).after_activity_id,
                      l_activities (indx).estimated_minutes,
                      l_activities (indx).difficulty_level);
   END;

   PROCEDURE set_workout (goal_id_in           IN INTEGER,
                          user_id_in           IN INTEGER,
                          topic_id_in             INTEGER,
                          week_number_in       IN INTEGER,
                          difficulty_id_in     IN INTEGER,
                          intensity_level_in   IN VARCHAR2,
                          goal_start_date_in   IN DATE DEFAULT NULL,
                          initial_setup_in     IN BOOLEAN DEFAULT FALSE)
   IS
      l_id              INTEGER;
      l_start           DATE := goal_start_date_in;
      l_workout         dg_workouts%ROWTYPE;
      l_workout_start   DATE;
      l_workout_end     DATE;

      FUNCTION this_workout
         RETURN dg_workouts%ROWTYPE
      IS
         l_workout   dg_workouts%ROWTYPE;
      BEGIN
         SELECT w.*
           INTO l_workout
           FROM dg_workouts w, dg_goals g
          WHERE     w.goal_id = g.goal_id
                AND w.week_number = week_number_in
                AND g.goal_id = goal_id_in;

         RETURN l_workout;
      EXCEPTION
         WHEN NO_DATA_FOUND
         THEN
            RETURN NULL;
      END;
   BEGIN
      l_workout := this_workout ();

      IF    l_workout.topic_id <> topic_id_in
         OR l_workout.difficulty_id <> difficulty_id_in
      THEN
         /* Just update and then only redo activity generation if changes.
            Note: going to ignore possible intensity level change for now!
         */

         UPDATE dg_workouts
            SET topic_id = topic_id_in, difficulty_id = difficulty_id_in
          WHERE workout_id = l_workout.workout_id;

         add_workout_activities (workout_id_in        => l_workout.workout_id,
                                 topic_id_in          => topic_id_in,
                                 intensity_level_in   => intensity_level_in,
                                 difficulty_id_in     => difficulty_id_in);
      ELSIF l_workout.workout_id IS NULL
      THEN
         IF goal_start_date_in IS NULL
         THEN
            SELECT start_date
              INTO l_start
              FROM dg_goals g
             WHERE g.goal_id = goal_id_in;
         END IF;

         get_week_start_ends (goal_start_in    => l_start,
                              week_in          => week_number_in,
                              week_start_out   => l_workout_start,
                              week_end_out     => l_workout_end);

         IF SYSDATE < l_workout_start OR initial_setup_in
         THEN
            INSERT INTO dg_workouts (goal_id,
                                     workout_name,
                                     topic_id,
                                     week_number,
                                     difficulty_id,
                                     start_date,
                                     end_date,
                                     user_id)
                 VALUES (goal_id_in,
                         'Goal ' || goal_id_in || ' Week ' || week_number_in,
                         topic_id_in,
                         week_number_in,
                         difficulty_id_in,
                         l_workout_start,
                         l_workout_end,
                         user_id_in)
              RETURNING workout_id
                   INTO l_id;

            add_workout_activities (
               workout_id_in        => l_id,
               topic_id_in          => topic_id_in,
               intensity_level_in   => intensity_level_in,
               difficulty_id_in     => difficulty_id_in);
         END IF;
      END IF;
   END;

   PROCEDURE remove_workout (workout_id_in IN INTEGER)
   IS
   BEGIN
      IF can_change_workout (workout_id_in)
      THEN
         DELETE FROM dg_workouts
               WHERE workout_id = workout_id_in;
      END IF;
   END;

   FUNCTION current_workout (user_id_in IN INTEGER)
      RETURN dg_workouts%ROWTYPE
   IS
      CURSOR nearest_workout_cur (
         date_in   IN DATE)
      IS
         SELECT wo.*
           FROM dg_workouts wo, dg_goals g
          WHERE     wo.goal_id = g.goal_id
                AND g.user_id = user_id_in
                AND date_in BETWEEN wo.start_date AND wo.end_date;

      l_workout   dg_workouts%ROWTYPE;
   BEGIN
      OPEN nearest_workout_cur (SYSDATE);

      FETCH nearest_workout_cur INTO l_workout;

      CLOSE nearest_workout_cur;

      IF l_workout.workout_id IS NULL
      THEN
         OPEN nearest_workout_cur (SYSDATE + 7);

         FETCH nearest_workout_cur INTO l_workout;

         CLOSE nearest_workout_cur;
      END IF;

      RETURN l_workout;
   END;

   FUNCTION current_workout_id (user_id_in IN INTEGER)
      RETURN INTEGER
   IS
      l_workout   dg_workouts%ROWTYPE;
   BEGIN
      l_workout := current_workout (user_id_in);

      RETURN l_workout.workout_id;
   END;

   FUNCTION workout_start_date (goal_id_in IN INTEGER, week_in IN INTEGER)
      RETURN DATE
   IS
      l_return   DATE;
   BEGIN
      SELECT wo.start_date
        INTO l_return
        FROM dg_workouts wo
       WHERE wo.goal_id = goal_id_in AND wo.week_number = week_in;

      RETURN l_return;
   EXCEPTION
      WHEN NO_DATA_FOUND
      THEN
         RETURN NULL;
   END;

   /* If week hasn't started or no activity in the week, can change */

   FUNCTION can_change_workout (workout_id_in IN INTEGER)
      RETURN BOOLEAN
   IS
      l_count   INTEGER;
      l_start   DATE;
   BEGIN
      SELECT wo.start_date
        INTO l_start
        FROM dg_workouts wo
       WHERE workout_id = workout_id_in;

      IF l_start > SYSDATE
      THEN
         RETURN TRUE;
      ELSE
         /* The week has started. Have you done anything yet? */

         SELECT COUNT (*)
           INTO l_count
           FROM dg_workout_activities wa
          WHERE     wa.completed_date IS NOT NULL
                AND wa.workout_id = workout_id_in;

         RETURN l_count = 0;
      END IF;
   END;

   /* Set up four workouts for the current/upcoming goal */

   FUNCTION resource_in_workout (
      activities_in   IN dg_workout_activities_nt,
      resource_in     IN ov.ov_resources_external%ROWTYPE)
      RETURN BOOLEAN
   IS
      l_dummy   VARCHAR2 (1);
   BEGIN
      FOR indx IN 1 .. activities_in.COUNT
      LOOP
         IF    activities_in (indx).url = resource_in.url
            OR activities_in (indx).activity_key_id =
                  resource_in.external_resource_key
         THEN
            RETURN TRUE;
         END IF;
      END LOOP;

      RETURN FALSE;
   END;

   FUNCTION resource_in_workout (workout_id_in   IN INTEGER,
                                 url_in          IN VARCHAR2)
      RETURN BOOLEAN
   IS
      l_dummy   VARCHAR2 (1);
   BEGIN
      SELECT 'x'
        INTO l_dummy
        FROM dg_workout_activities wa
       WHERE wa.workout_id = workout_id_in AND url = url_in;

      RETURN TRUE;
   EXCEPTION
      WHEN NO_DATA_FOUND
      THEN
         RETURN FALSE;
   END;

   FUNCTION resource_in_workout (activities_in   IN dg_workout_activities_nt,
                                 url_in          IN VARCHAR2)
      RETURN BOOLEAN
   IS
   BEGIN
      FOR indx IN 1 .. activities_in.COUNT
      LOOP
         IF activities_in (indx).url = url_in
         THEN
            RETURN TRUE;
         END IF;
      END LOOP;

      RETURN FALSE;
   END;

   FUNCTION next_workout_activity (user_id_in IN INTEGER)
      RETURN dg_workout_activities%ROWTYPE
   IS
      CURSOR activities_cur
      IS
           SELECT wa.*
             FROM dg_workout_activities wa, dg_activity_types tp
            WHERE     wa.workout_id = current_workout_id (user_id_in)
                  AND wa.activity_type_id = tp.activity_type_id
                  AND wa.completed_date IS NULL
         ORDER BY activity_sequence ASC;

      l_Return   dg_workout_activities%ROWTYPE;
   BEGIN
      OPEN activities_cur;

      FETCH activities_cur INTO l_return;

      CLOSE activities_cur;

      RETURN l_return;
   END;

   FUNCTION next_workout_activity_url (user_id_in IN INTEGER)
      RETURN VARCHAR2
   IS
      l_wa   dg_workout_activities%ROWTYPE;
   BEGIN
      l_wa := next_workout_activity (user_id_in);

      RETURN l_wa.url;
   END;

   FUNCTION next_workout_activity_nane (user_id_in IN INTEGER)
      RETURN VARCHAR2
   IS
      l_wa   dg_workout_activities%ROWTYPE;
   BEGIN
      l_wa := next_workout_activity (user_id_in);

      RETURN l_wa.activity_name;
   END;

   FUNCTION completed_count (workout_id_in IN INTEGER)
      RETURN INTEGER
   IS
      l_return   INTEGER;
   BEGIN
      SELECT COUNT (*)
        INTO l_return
        FROM dg_workout_activities
       WHERE workout_id = workout_id_in AND completed_date IS NOT NULL;

      RETURN l_return;
   END;

   FUNCTION not_done_count (workout_id_in IN INTEGER)
      RETURN INTEGER
   IS
      l_return   INTEGER;
   BEGIN
      SELECT COUNT (*)
        INTO l_return
        FROM dg_workout_activities
       WHERE workout_id = workout_id_in AND completed_date IS NULL;

      RETURN l_return;
   END;

   /* How many are left? */

   FUNCTION unplayed_quizzes_count (user_id_IN INTEGER)
      RETURN INTEGER
   IS
      l_count   INTEGER;
   BEGIN
      SELECT COUNT (*)
        INTO l_count
        FROM (SELECT question_id FROM dg_training_questions_v
              MINUS
              SELECT z.question_id
                FROM qdb_quizzes z, qdb_compev_answers a
               WHERE     a.user_id = user_id_in
                     AND z.comp_event_id = a.comp_event_id
                     AND a.ended_on IS NOT NULL);

      RETURN l_count;
   END;

   FUNCTION minutes_for_level (intensity_level_in IN VARCHAR2)
      RETURN PLS_INTEGER
   IS
   BEGIN
      RETURN CASE intensity_level_in
                WHEN c_intensity_low THEN 60
                WHEN c_intensity_medium THEN 180
                WHEN c_intensity_high THEN 300
             END;
   END;

   FUNCTION act_type_from_res_usage (usage_in IN VARCHAR2)
      RETURN INTEGER
   IS
      /* Need to convert from the many CH usages to the "higher level"
      activity types. We have:
      Train - non-competitive quizzes
      Compete - competitive quizzes
      Test - take a test (multi-question quiz)
      Challenge - write code to solve a problem
      Read - read a blog post, whitepaper, etc.
      Watch - watch a video
      Execute - run a script or utility, go through a tutorial
      */
      l_title    dg_activity_types.title%TYPE;
      l_return   dg_activity_types.activity_type_id%TYPE;
   BEGIN
      l_title :=
         CASE usage_in
            WHEN 'BLOG' THEN 'Read'
            WHEN 'Article' THEN 'Read'
            WHEN 'White paper' THEN 'Read'
            WHEN 'OBE' THEN 'Execute'
            WHEN 'Tutorial' THEN 'Execute'
            WHEN 'Demo' THEN 'Execute'
            WHEN 'Demonstration' THEN 'Execute'
            WHEN 'Quiz' THEN 'Train'
            WHEN 'Challenge' THEN 'Challenge'
            WHEN 'Test' THEN 'Test'
            WHEN 'Video' THEN 'Watch'
            ELSE 'Read'
         END;

      SELECT activity_type_id
        INTO l_return
        FROM dg_activity_types
       WHERE title = l_title;

      RETURN l_return;
   END;

   FUNCTION workout_start (goal_start_in IN DATE, week_in IN INTEGER)
      RETURN DATE
   IS
   BEGIN
      RETURN qdb_utilities.start_of_day (goal_start_in + (week_in - 1) * 7);
   END;

   FUNCTION workout_end (goal_start_in IN DATE, week_in IN INTEGER)
      RETURN DATE
   IS
   BEGIN
      RETURN qdb_utilities.end_of_day (
                workout_start (goal_start_in, week_in) + 6);
   END;

   PROCEDURE get_week_start_ends (goal_start_in    IN     DATE,
                                  week_in          IN     INTEGER,
                                  week_start_out      OUT DATE,
                                  week_end_out        OUT DATE)
   IS
   BEGIN
      week_start_out := workout_start (goal_start_in, week_in);
      week_end_out := workout_end (goal_start_in, week_in);
   END;

   PROCEDURE get_suggested_topics (user_id_in      IN     INTEGER,
                                   topic1_id_out      OUT INTEGER,
                                   topic2_id_out      OUT INTEGER,
                                   topic3_id_out      OUT INTEGER,
                                   topic4_id_out      OUT INTEGER)
   IS
      CURSOR topics_cur
      IS
         WITH user_domains_count
              AS (SELECT COUNT (*) dcount
                    FROM qdb_user_domains
                   WHERE user_id = user_id_in),
              user_domains
              AS (SELECT domain_id
                    FROM qdb_user_domains, user_domains_count
                   WHERE     user_id = user_id_in
                         AND user_domains_count.dcount > 0
                  UNION
                  SELECT domain_id
                    FROM qdb_public_domains_v, user_domains_count
                   WHERE user_domains_count.dcount = 0)
           SELECT topic_id
             FROM dg_suggested_topics
            WHERE domain_id IN (SELECT domain_id FROM user_domains)
         ORDER BY priority_order, domain_id;

      topic_rec   topics_cur%ROWTYPE;
      l_topics    DBMS_SQL.number_table;
   BEGIN
      OPEN topics_cur;

      FETCH topics_cur INTO topic_rec;

      l_topics (1) := topic_rec.topic_id;

      FETCH topics_cur INTO topic_rec;

      IF topics_cur%NOTFOUND
      THEN
         l_topics (2) := l_topics (1);
         l_topics (3) := l_topics (1);
         l_topics (4) := l_topics (1);
      ELSE
         l_topics (2) := topic_rec.topic_id;

         FETCH topics_cur INTO topic_rec;

         IF topics_cur%NOTFOUND
         THEN
            l_topics (3) := l_topics (1);
            l_topics (4) := l_topics (2);
         ELSE
            l_topics (3) := topic_rec.topic_id;

            FETCH topics_cur INTO topic_rec;

            IF topics_cur%NOTFOUND
            THEN
               l_topics (4) := l_topics (1);
            ELSE
               l_topics (4) := topic_rec.topic_id;
            END IF;
         END IF;
      END IF;

      topic1_id_out := l_topics (1);
      topic2_id_out := l_topics (2);
      topic3_id_out := l_topics (3);
      topic4_id_out := l_topics (4);
   END;

   PROCEDURE set_up_goal (
      user_id_in                IN     INTEGER,
      target_pct_correct_in     IN     INTEGER,
      target_min_quizzes_in     IN     INTEGER,
      target_daily_ranking_in   IN     INTEGER,
      intensity_level_in        IN     VARCHAR2,
      play_daily_db_quiz_in     IN     VARCHAR2,
      start_date_in             IN     DATE DEFAULT NULL,
      end_date_in               IN     DATE DEFAULT NULL,
      topic1_id_in              IN     INTEGER,
      topic2_id_in              IN     INTEGER DEFAULT NULL,
      topic3_id_in              IN     INTEGER DEFAULT NULL,
      topic4_id_in              IN     INTEGER DEFAULT NULL,
      topic1_difficulty_id_in   IN     INTEGER,
      topic2_difficulty_id_in   IN     INTEGER DEFAULT NULL,
      topic3_difficulty_id_in   IN     INTEGER DEFAULT NULL,
      topic4_difficulty_id_in   IN     INTEGER DEFAULT NULL,
      goal_id_out                  OUT INTEGER)
   IS
      l_id             INTEGER;
      l_topics         DBMS_SQL.number_table;
      l_difficulties   DBMS_SQL.number_table;
   BEGIN
      insert_goal (user_id_in                => user_id_in,
                   target_pct_correct_in     => target_pct_correct_in,
                   target_min_quizzes_in     => target_min_quizzes_in,
                   target_daily_ranking_in   => target_daily_ranking_in,
                   intensity_level_in        => intensity_level_in,
                   play_daily_db_quiz_in     => play_daily_db_quiz_in,
                   start_date_in             => start_date_in,
                   end_date_in               => end_date_in,
                   goal_id_out               => l_id);

      IF topic1_id_in IS NOT NULL
      THEN
         l_topics (l_topics.COUNT + 1) := topic1_id_in;
         l_difficulties (l_topics.COUNT) := topic1_difficulty_id_in;
      END IF;

      IF topic2_id_in IS NOT NULL
      THEN
         l_topics (l_topics.COUNT + 1) := topic2_id_in;
         l_difficulties (l_topics.COUNT) := topic2_difficulty_id_in;
      END IF;

      IF topic3_id_in IS NOT NULL
      THEN
         l_topics (l_topics.COUNT + 1) := topic3_id_in;
         l_difficulties (l_topics.COUNT) := topic3_difficulty_id_in;
      END IF;

      IF topic4_id_in IS NOT NULL
      THEN
         l_topics (l_topics.COUNT + 1) := topic4_id_in;
         l_difficulties (l_topics.COUNT) := topic4_difficulty_id_in;
      END IF;

      CASE l_topics.COUNT
         WHEN 1
         THEN
            l_topics (2) := l_topics (1);
            l_difficulties (2) := l_difficulties (1);
            l_topics (3) := l_topics (1);
            l_difficulties (3) := l_difficulties (1);
            l_topics (4) := l_topics (1);
            l_difficulties (4) := l_difficulties (1);
         WHEN 2
         THEN
            l_topics (3) := l_topics (1);
            l_difficulties (3) := l_difficulties (1);
            l_topics (4) := l_topics (2);
            l_difficulties (4) := l_difficulties (2);
         WHEN 3
         THEN
            l_topics (4) := l_topics (1);
            l_difficulties (4) := l_difficulties (1);
         WHEN 4
         THEN
            /* Four selected, all filled, done */
            NULL;
         WHEN 0
         THEN
            /* No topics selected. Nothing happens. */
            l_topics.DELETE;
      END CASE;

      FOR indx IN 1 .. l_topics.COUNT
      LOOP
         set_workout (goal_id_in           => l_id,
                      user_id_in           => user_id_in,
                      topic_id_in          => l_topics (indx),
                      week_number_in       => indx,
                      intensity_level_in   => intensity_level_in,
                      difficulty_id_in     => l_difficulties (indx),
                      initial_setup_in     => TRUE);
      END LOOP;

      goal_id_out := l_id;
   EXCEPTION
      WHEN OTHERS
      THEN
         qdb_utilities.syslog_error ('set_up_goal',
                                     'Failed goal ID',
                                     l_id,
                                     raise_error_in   => FALSE);
         RAISE;
   END;

   /* Use a nice solid set of defaults */

   PROCEDURE set_up_goal (
      user_id_in                IN     INTEGER,
      target_pct_correct_in     IN     INTEGER DEFAULT NULL,
      target_min_quizzes_in     IN     INTEGER DEFAULT NULL,
      target_daily_ranking_in   IN     INTEGER DEFAULT NULL,
      intensity_level_in        IN     VARCHAR2 DEFAULT NULL,
      play_daily_db_quiz_in     IN     VARCHAR2 DEFAULT NULL,
      start_date_in             IN     DATE DEFAULT NULL,
      end_date_in               IN     DATE DEFAULT NULL,
      goal_id_out                  OUT INTEGER)
   IS
      l_topics   DBMS_SQL.number_table;
   BEGIN
      dg_workout_mgr.get_suggested_topics (user_id_in      => user_id_in,
                                           topic1_id_out   => l_topics (1),
                                           topic2_id_out   => l_topics (2),
                                           topic3_id_out   => l_topics (3),
                                           topic4_id_out   => l_topics (4));
      /* Always intermediate difficulty */
      set_up_goal (
         user_id_in                => user_id_in,
         target_pct_correct_in     => NVL (target_pct_correct_in, 70),
         target_min_quizzes_in     => NVL (target_min_quizzes_in, 10),
         target_daily_ranking_in   => NVL (target_daily_ranking_in, 100),
         intensity_level_in        => NVL (intensity_level_in,
                                           dg_workout_mgr.c_intensity_low),
         play_daily_db_quiz_in     => 'Y',
         start_date_in             => start_date_in,
         end_date_in               => end_date_in,
         topic1_id_in              => l_topics (1),
         topic2_id_in              => l_topics (2),
         topic3_id_in              => l_topics (3),
         topic4_id_in              => l_topics (4),
         topic1_difficulty_id_in   => 3,
         topic2_difficulty_id_in   => 3,
         topic3_difficulty_id_in   => 3,
         topic4_difficulty_id_in   => 3,
         goal_id_out               => goal_id_out);
   END;

   PROCEDURE mark_activity_completed (
      workout_activity_id_in   IN INTEGER,
      minutes_to_complete_in   IN NUMBER,
      user_rating_in           IN INTEGER DEFAULT NULL)
   IS
      l_wo_id     INTEGER;
      l_user_id   INTEGER;
   BEGIN
   select completed_date into l_date
   from dg_workout_activities
            
          WHERE workout_activity_id = workout_activity_id_in;
          
         UPDATE dg_workout_activities
            SET completed_date = SYSDATE,
                actual_minutes = minutes_to_complete_in,
                user_rating = user_rating_in
          WHERE workout_activity_id = workout_activity_id_in
      RETURNING workout_id
           INTO l_wo_id;

     /* Newly completed, so give points */
     if l_date is null then
     SELECT user_id
        INTO l_user_id
        FROM dg_workouts
       WHERE workout_id = l_wo_id;

      qdb_point_mgr.add_points (
         user_id_in        => l_user_id,
         activity_id_in    => qdb_point_mgr.activity_id_from_abbrev ('WOACT'),
         activity_key_in   => workout_activity_id_in);end if;
   END;

   /* Look up URL in current workout */

   PROCEDURE mark_activity_completed (
      user_id_in               IN INTEGER,
      url_in                   IN VARCHAR2,
      minutes_to_complete_in   IN NUMBER,
      user_rating_in           IN INTEGER DEFAULT NULL)
   IS
   l_id integer;
   BEGIN
   SELECT workout_activity_id into l_id
                   FROM dg_workout_activities w
                  WHERE     workout_id = current_workout_id (user_id_in)
                        AND url = url_in;
                     mark_activity_completed (
      workout_activity_id_in  => l_id,
      minutes_to_complete_in   =>minutes_to_complete_in,
      user_rating_in    =>user_rating_in);
   END;

   /* Special tie-in for comp event */

   PROCEDURE mark_quiz_completed (user_id_in         IN INTEGER,
                                  comp_event_id_in   IN INTEGER)
   IS
      l_minutes       NUMBER;
      l_question_id   INTEGER;
      l_ended_on      DATE;
      l_id integer;
   BEGIN
      SELECT ROUND ( (eva.ended_on - eva.started_on) * 86400 / 60, 1),
             q.question_id,
             eva.ended_on
        INTO l_minutes, l_question_id, l_ended_on
        FROM qdb_compev_answers eva, qdb_quizzes q
       WHERE     eva.user_id = user_id_in
             AND eva.comp_event_id = comp_event_id_in
             AND eva.comp_event_id = q.comp_event_id
             AND q.answer_sequence = 1;

      /* The activity_key_id is the question ID.
      If not part of the workout, no harm done.
      */
      begin
      
   SELECT workout_activity_id into l_id
                   FROM dg_workout_activities w
                  WHERE     workout_id = current_workout_id (user_id_in)
                       AND activity_key_id = l_question_id;
                       
                     mark_activity_completed (
      workout_activity_id_in  => l_id,
      minutes_to_complete_in   =>l_minutes,
      user_rating_in    =>NULL);exception when no_data_found
      then
      /* Just skip this step */
      null;
      end;
   END;

   FUNCTION activity_id_for_ch_resource_id (workout_id_in      IN INTEGER,
                                            ch_content_id_in   IN INTEGER)
      RETURN INTEGER
   IS
      l_return   INTEGER;
   BEGIN
      SELECT workout_activity_id
        INTO l_return
        FROM dg_workout_activities
       WHERE workout_id = workout_id_in AND ch_content_id = ch_content_id_in;

      RETURN l_return;
   EXCEPTION
      WHEN NO_DATA_FOUND
      THEN
         RETURN NULL;
   END;

   PROCEDURE put_activity_after (workout_activity_id_in   IN INTEGER,
                                 after_activity_id_in     IN INTEGER)
   IS
   BEGIN
      UPDATE dg_workout_activities
         SET after_activity_id = after_activity_id_in
       WHERE workout_activity_id = workout_activity_id_in;
   END;

   PROCEDURE ins_activity (
      workout_id_in            IN dg_workout_activities.workout_id%TYPE,
      activity_type_id_in         dg_workout_activities.activity_type_id%TYPE,
      ch_content_id_in            dg_workout_activities.ch_content_id%TYPE,
      activity_key_id_in          dg_workout_activities.activity_key_id%TYPE,
      activity_sequence_in        dg_workout_activities.activity_sequence%TYPE,
      activity_name               dg_workout_activities.activity_name%TYPE,
      url_in                      dg_workout_activities.url%TYPE,
      activity_source_in          dg_workout_activities.activity_source%TYPE DEFAULT NULL,
      activity_source_url_in      dg_workout_activities.activity_source_url%TYPE DEFAULT NULL,
      after_activity_id_in        dg_workout_activities.after_activity_id%TYPE DEFAULT NULL,
      difficulty_level_in         dg_workout_activities.difficulty_level%TYPE DEFAULT NULL,
      estimated_minutes_in        dg_workout_activities.estimated_minutes%TYPE DEFAULT NULL,
      completed_date_in           dg_workout_activities.completed_date%TYPE DEFAULT 'N',
      creation_type_in            dg_workout_activities.creation_type%TYPE DEFAULT c_generated_workout)
   IS
   BEGIN
      raise_application_error (-20000, 'ins_activity not implemented');
   END;

   FUNCTION random_question_query (
      user_id_in         IN INTEGER,
      domain_id_in       IN INTEGER DEFAULT NULL,
      topic_id_in        IN INTEGER DEFAULT NULL,
      difficulty_id_in   IN INTEGER DEFAULT NULL,
      filter_in          IN VARCHAR2 DEFAULT NULL,
      unplayed_in        IN BOOLEAN DEFAULT TRUE)
      RETURN VARCHAR2
   IS
      /* If domain set, convert to root topic and use that.
      If topic is set, ignore domain.
      If neither is set, then filter by user domains.
      */
      l_query          qdb_config.maxvarchar2
         := 'SELECT question_id FROM (SELECT question_id FROM dg_training_questions_v tq';
      l_domain_count   INTEGER;
      l_cur            SYS_REFCURSOR;
      l_selected       INTEGER;
   BEGIN
      SELECT COUNT (*)
        INTO l_domain_count
        FROM qdb_user_domains
       WHERE user_id = user_id_in;

      l_query :=
            l_query
         || CASE
               WHEN domain_id_in IS NULL AND topic_id_in IS NULL
               THEN
                  CASE
                     WHEN l_domain_count > 0
                     THEN
                           ' WHERE domain_id IN (SELECT domain_id FROM qdb_user_domains WHERE user_id = '
                        || user_id_in
                        || ')'
                  END
               WHEN topic_id_in IS NOT NULL
               THEN
                     ' WHERE tq.topic_id IN (SELECT t2.topic_id from qdb_topics t2'
                  || ' CONNECT BY prior t2.topic_id = t2.parent_topic_id START WITH t2.topic_id = '
                  || topic_id_in
                  || ')'
               WHEN domain_id_in IS NOT NULL
               THEN
                  ' WHERE domain_id = ' || domain_id_in
            END
         || CASE
               WHEN difficulty_id_in IS NOT NULL
               THEN
                  ' AND difficulty_id = ' || difficulty_id_in
            END
         || CASE
               WHEN filter_in IS NOT NULL
               THEN
                  /* Find questions with topics whose names match. */
                  ' AND UPPER (tq.topic) LIKE ''%'' || upper (filter_in) || ''%'''
            END;
      l_query :=
            l_query
         || CASE
               WHEN unplayed_in
               THEN
                     'MINUS SELECT z.question_id FROM qdb_quizzes z, qdb_compev_answers a'
                  || ' WHERE a.user_id = '
                  || user_id_in
                  || 'AND z.comp_event_id = a.comp_event_id'
            END
         || ') ORDER BY DBMS_RANDOM.VALUE ()';

      RETURN l_query;
   END;

   FUNCTION random_question (user_id_in     IN INTEGER,
                             domain_id_in   IN INTEGER DEFAULT NULL,
                             topic_id_in    IN INTEGER DEFAULT NULL)
      RETURN INTEGER
   IS
      l_cur        SYS_REFCURSOR;
      l_selected   INTEGER;
   BEGIN
      OPEN l_cur FOR
         random_question_query (user_id_in,
                                domain_id_in,
                                topic_id_in,
                                unplayed_in   => TRUE);

      FETCH l_cur INTO l_selected;

      CLOSE l_cur;

      IF l_selected IS NULL
      THEN
         /* Try again without restriction on already played. */
         OPEN l_cur FOR
            random_question_query (user_id_in,
                                   domain_id_in,
                                   topic_id_in,
                                   unplayed_in   => FALSE);

         FETCH l_cur INTO l_selected;

         CLOSE l_cur;
      END IF;

      RETURN l_selected;
   END;

   /* Feed question IDs back to page 3 for further computation and display */

   FUNCTION random_questions (user_id_in          IN INTEGER,
                              domain_id_in        IN INTEGER DEFAULT NULL,
                              topic_id_in         IN INTEGER DEFAULT NULL,
                              difficulty_id_in    IN INTEGER DEFAULT NULL,
                              filter_in           IN VARCHAR2 DEFAULT NULL,
                              question_limit_in   IN INTEGER DEFAULT 20)
      RETURN numbers_nt
   IS
      /* Return a "core" set of randomly selected questions to be displayed
      to players who want to take a quiz on a given topic.
      Questions never answered take priority over those taken before.
      */
      l_cur                SYS_REFCURSOR;
      l_questions          numbers_nt := numbers_nt ();
      l_questions_played   numbers_nt := numbers_nt ();
   BEGIN
      /* First check for unplayed quizzes */
      OPEN l_cur FOR
            'SELECT * FROM ('
         || random_question_query (user_id_in,
                                   domain_id_in,
                                   topic_id_in,
                                   difficulty_id_in   => difficulty_id_in,
                                   unplayed_in        => TRUE)
         || ' ) FETCH FIRST '
         || question_limit_in
         || ' ROWS ONLY';

      FETCH l_cur BULK COLLECT INTO l_questions;

      CLOSE l_cur;

      IF l_questions.COUNT < question_limit_in
      THEN
         OPEN l_cur FOR
               'SELECT * FROM ('
            || random_question_query (user_id_in,
                                      domain_id_in,
                                      topic_id_in,
                                      difficulty_id_in   => difficulty_id_in,
                                      unplayed_in        => FALSE)
            || ' ) FETCH FIRST '
            || question_limit_in
            || ' ROWS ONLY';

         /* Pick up N already played to the max number */

         FETCH l_cur BULK COLLECT INTO l_questions_played;

         CLOSE l_cur;
      END IF;

      /* use union to pull over any questions in "played" */

      RETURN l_questions MULTISET UNION DISTINCT l_questions_played;
   END;

   /* USE WITH CAUTION! For user testing only. */

   PROCEDURE reset_users (user_id_in IN INTEGER DEFAULT NULL)
   IS
   BEGIN
      qdb_utilities.syslog_assert (
         user_id_in IS NOT NULL,
         'Resetting of all users has been disabled.');

      DELETE FROM dg_workout_activities
            WHERE workout_id IN (SELECT workout_id
                                   FROM dg_goals g, dg_workouts wo
                                  WHERE     (   g.user_id = user_id_in
                                             OR user_id_in IS NULL)
                                        AND g.goal_id = wo.goal_id);

      DELETE FROM dg_workouts
            WHERE goal_id IN (SELECT goal_id
                                FROM dg_goals g
                               WHERE    g.user_id = user_id_in
                                     OR user_id_in IS NULL);

      DELETE FROM dg_goals g
            WHERE g.user_id = user_id_in OR user_id_in IS NULL;

      IF user_id_in IS NULL
      THEN
         UPDATE qdb_users
            SET saw_dg_welcome = qdb_config.c_no;
      ELSE
         UPDATE qdb_users
            SET saw_dg_welcome = qdb_config.c_no
          WHERE user_id = user_id_in;
      END IF;
   END;

   FUNCTION unplayed_compete_count (user_id_in IN INTEGER)
      RETURN INTEGER
   IS
      l_count   INTEGER;
   BEGIN
      SELECT COUNT (*)
        INTO l_count
        FROM qdb_approved_comp_events_v a
       WHERE     a.comp_type_text = 'COMPETE'
             AND SYSDATE BETWEEN a.start_date AND a.end_date
             AND comp_event_id NOT IN (SELECT comp_event_id
                                         FROM qdb_compev_answers
                                        WHERE     user_id = user_id_in
                                              AND ended_on IS NOT NULL)
             AND a.domain_id IN (SELECT domain_id
                                   FROM qdb_user_domains
                                  WHERE user_id = user_id_in);

      RETURN l_count;
   END;

   FUNCTION is_dev_gym_activity (workout_activity_id_in IN INTEGER)
      RETURN BOOLEAN
   IS
      l_dummy   CHAR (1);
   BEGIN
      SELECT 'x'
        INTO l_dummy
        FROM dg_workout_activities
       WHERE     workout_activity_id = workout_activity_id_in
             AND url LIKE
                       'f?p='
                    || qdb_hub_integration_mgr.c_app_source_dev_gym
                    || '%';

      RETURN TRUE;
   EXCEPTION
      WHEN NO_DATA_FOUND
      THEN
         RETURN FALSE;
   END;

   /* Custom workouts */

   PROCEDURE create_custom_workout (
      workout_name_in    IN VARCHAR2,
      description_in     IN VARCHAR2,
      author_id_in       IN INTEGER,
      topic_id_in        IN INTEGER,
      start_date_in      IN DATE DEFAULT NULL,
      end_date_in        IN DATE DEFAULT NULL,
      difficulty_id_in   IN INTEGER DEFAULT NULL)
   IS
   BEGIN
      INSERT INTO DG_WORKOUTS (WORKOUT_NAME,
                               DESCRIPTION,
                               AUTHOR_ID,
                               TOPIC_ID,
                               START_DATE,
                               END_DATE,
                               DIFFICULTY_ID,
                               WORKOUT_TYPE)
           VALUES (workout_name_in,
                   description_in,
                   author_id_in,
                   topic_id_in,
                   start_date_in,
                   end_date_in,
                   difficulty_id_in,
                   c_custom_workout);
   END;
END;
/