/* Very basic error logging */

CREATE TABLE errlog
(
   err_seq         NUMBER GENERATED ALWAYS AS IDENTITY,
   err_code        INTEGER,
   err_message     VARCHAR2 (4000),
   err_backtrace   VARCHAR2 (4000),
   app_info        VARCHAR2 (4000)
)
/

CREATE OR REPLACE PROCEDURE log_error (app_info_in IN VARCHAR2)
IS
   l_code   INTEGER := SQLCODE;
   PRAGMA AUTONOMOUS_TRANSACTION;
BEGIN
   INSERT INTO errlog (err_code,
                       err_message,
                       err_backtrace,
                       app_info)
        VALUES (l_code,
                DBMS_UTILITY.format_error_stack,
                DBMS_UTILITY.format_error_backtrace,
                app_info_in);

   COMMIT;
END;
/

CREATE OR REPLACE PROCEDURE innermost
IS
   l_divisor   NUMBER := 0;
BEGIN
   DBMS_OUTPUT.put_line (100 / l_divisor);
END;
/

/* Only in outermost block */

CREATE OR REPLACE PROCEDURE innermost
IS
   l_divisor   NUMBER := 0;
BEGIN
   DBMS_OUTPUT.put_line (100 / l_divisor);
END;
/

CREATE OR REPLACE PROCEDURE inner
IS
BEGIN
   innermost;
END;
/

CREATE OR REPLACE PROCEDURE uppermost
IS
BEGIN
   inner;
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (app_info_in => 'Uppermost Handler');
END;
/

BEGIN
   uppermost;
END;
/

  SELECT *
    FROM errlog
ORDER BY err_seq
/

TRUNCATE TABLE errlog
/

/* Handle in lowest block */

CREATE OR REPLACE PROCEDURE innermost
IS
   l_divisor   NUMBER := 0;
BEGIN
   DBMS_OUTPUT.put_line (100 / l_divisor);
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (
         app_info_in   => 'Innermost Handler - Value of divisor: ' || l_divisor);
      RAISE;
END;
/

CREATE OR REPLACE PROCEDURE inner
IS
BEGIN
   innermost;
END;
/

CREATE OR REPLACE PROCEDURE uppermost
IS
BEGIN
   inner;
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (app_info_in => 'Uppermost Handler');
END;
/

BEGIN
   uppermost;
END;
/

  SELECT *
    FROM errlog
ORDER BY err_seq
/

TRUNCATE TABLE errlog
/

/* Handle in every block */

CREATE OR REPLACE PROCEDURE innermost
IS
   l_divisor   NUMBER := 0;
BEGIN
   DBMS_OUTPUT.put_line (100 / l_divisor);
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (
         app_info_in   => 'Innermost Handler - Value of divisor: ' || l_divisor);
      RAISE;
END;
/

CREATE OR REPLACE PROCEDURE inner
IS
BEGIN
   innermost;
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (app_info_in => 'Inner Handler');
      RAISE;
END;
/

CREATE OR REPLACE PROCEDURE uppermost
IS
BEGIN
   inner;
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (app_info_in => 'Uppermost Handler');
END;
/

BEGIN
   uppermost;
END;
/

  SELECT *
    FROM errlog
ORDER BY err_seq
/

TRUNCATE TABLE errlog
/

/* Avoid repetitive and misleading logging */

CREATE OR REPLACE PROCEDURE log_error (app_info_in   IN VARCHAR2,
                                       reraise_in    IN BOOLEAN DEFAULT FALSE)
IS
   c_reraise_error_code constant integer := -20177;
   l_code   INTEGER := SQLCODE;
   PRAGMA AUTONOMOUS_TRANSACTION;
BEGIN
   IF l_code <> c_reraise_error_code
   THEN
      INSERT INTO errlog (err_code,
                          err_message,
                          err_backtrace,
                          app_info)
           VALUES (l_code,
                   DBMS_UTILITY.format_error_stack,
                   DBMS_UTILITY.format_error_backtrace,
                   app_info_in);

      COMMIT;
   END IF;

   IF reraise_in
   THEN
      raise_application_error (
         c_reraise_error_code,
         'Contact application administrator for assistance.');
   END IF;
END;
/

CREATE OR REPLACE PROCEDURE innermost
IS
   l_divisor   NUMBER := 0;
BEGIN
   DBMS_OUTPUT.put_line (100 / l_divisor);
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (
         app_info_in   => 'Innermost Handler - Value of divisor: ' || l_divisor,
         reraise_in    => TRUE);
END;
/

CREATE OR REPLACE PROCEDURE inner
IS
BEGIN
   innermost;
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (app_info_in => 'Inner Handler', reraise_in => TRUE);
END;
/

CREATE OR REPLACE PROCEDURE uppermost
IS
BEGIN
   inner;
EXCEPTION
   WHEN OTHERS
   THEN
      log_error (app_info_in => 'Uppermost Handler');
END;
/

BEGIN
   uppermost;
END;
/

  SELECT *
    FROM errlog
ORDER BY err_seq
/

TRUNCATE TABLE errlog
/

/* More info during development - "I don't want to have to query the error log every time" */

ALTER SESSION SET PLSQL_CCFLAGS="indevelopment:TRUE"
/

CREATE OR REPLACE PROCEDURE log_error (app_info_in   IN VARCHAR2,
                                       reraise_in    IN BOOLEAN DEFAULT FALSE)
IS
   c_reraise_error_code constant integer := -20177;
   l_code   INTEGER := SQLCODE;
   PRAGMA AUTONOMOUS_TRANSACTION;
BEGIN
   $IF $$in_development
   $THEN
      DBMS_OUTPUT.put_line (
         'Log error at ' || TO_CHAR (SYSDATE, 'YYYY-MM-DD HH24:MI:SS'));

      IF app_info_in IS NOT NULL
      THEN
         DBMS_OUTPUT.put_line ('App info: ' || app_info_in);
      END IF;

      DBMS_OUTPUT.put_line ('Error code: ' || l_code);
      DBMS_OUTPUT.put_line (
         'Error message: ' || DBMS_UTILITY.format_error_stack);
      DBMS_OUTPUT.put_line (
         'Error backtrace: ' || DBMS_UTILITY.format_error_backtrace);

      IF reraise_in
      THEN
         raise_application_error (
            -20177,
            'Error in Dev: ' || DBMS_UTILITY.format_error_stack);
      END IF;
   $ELSE
      IF l_code <> c_reraise_error_code
      THEN
         INSERT INTO errlog (err_code,
                             err_message,
                             err_backtrace,
                             app_info)
              VALUES (l_code,
                      DBMS_UTILITY.format_error_stack,
                      DBMS_UTILITY.format_error_backtrace,
                      app_info_in);

         COMMIT;
      END IF;

      IF reraise_in
      THEN
         raise_application_error (
            c_reraise_error_code,
            'Contact application administrator for assistance.');
      END IF;
   $END
END;
/

BEGIN
   uppermost;
END;
/

  SELECT *
    FROM errlog
ORDER BY err_seq
/
