DROP TABLE employees_m2
/

CREATE TABLE employees_m2
AS
   SELECT * FROM employees
/

CREATE OR REPLACE FUNCTION table_count (
   owner_in        IN all_tables.owner%TYPE
 , table_name_in   IN all_tables.table_name%TYPE
 , where_in        IN VARCHAR2 DEFAULT NULL)
   RETURN PLS_INTEGER
   AUTHID CURRENT_USER
IS
   l_return       PLS_INTEGER;
BEGIN
   EXECUTE IMMEDIATE
      'SELECT COUNT(*) FROM ' || owner_in || '.' || table_name_in ||
      CASE WHEN where_in IS NOT NULL THEN ' WHERE ' || where_in END
      INTO l_return;

   RETURN l_return;
END;
/

/* Trying to bind a column name... */

CREATE OR REPLACE PROCEDURE set_to_10000 (col_in             IN VARCHAR2
                                        , department_id_in   IN PLS_INTEGER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees_m2 SET :colname = 10000 
        WHERE department_id = :dept';
BEGIN
   EXECUTE IMMEDIATE l_update USING col_in, department_id_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/

BEGIN
   set_to_10000 ('salary', 50);
   /* ORA-01747: invalid user.table.column, table.column, or column specification */
END;
/

/* Single bind variable */

CREATE OR REPLACE PROCEDURE set_to_10000 (col_in             IN VARCHAR2
                                        , department_id_in   IN PLS_INTEGER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees_m2 SET '
         || col_in
         || ' = 10000 
        WHERE department_id = :dept';
BEGIN
   EXECUTE IMMEDIATE l_update USING department_id_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/

BEGIN
   DBMS_OUTPUT.put_line (
      'Before '
      || table_count (USER
                    , 'employees_m2'
                    , 'department_id = 50 AND salary = 10000'));
   set_to_10000 ('salary', 50);
   DBMS_OUTPUT.put_line (
      'After '
      || table_count (USER
                    , 'employees_m2'
                    , 'department_id = 50 AND salary = 10000'));
   ROLLBACK;
END;
/

/* Multiple bind variables */

CREATE OR REPLACE PROCEDURE updnumval (col_in             IN VARCHAR2
                                     , department_id_in   IN PLS_INTEGER
                                     , val_in             IN NUMBER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees_m2 SET '
         || col_in
         || ' = :val 
        WHERE department_id = :dept';
BEGIN
   EXECUTE IMMEDIATE l_update USING val_in, department_id_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/

BEGIN
   DBMS_OUTPUT.put_line (
      'Before '
      || table_count (USER
                    , 'employees_m2'
                    , 'department_id = 50 AND salary = 10000'));
   updnumval ('salary', 50, 10000);
   DBMS_OUTPUT.put_line (
      'After '
      || table_count (USER
                    , 'employees_m2'
                    , 'department_id = 50 AND salary = 10000'));
   ROLLBACK;
END;
/

/* Reference same bind variable more than once.

   Must supply value both times.
*/

CREATE OR REPLACE PROCEDURE updnumval (col_in             IN VARCHAR2
                                     , department_id_in   IN PLS_INTEGER
                                     , val_in             IN NUMBER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees_m2 SET '
         || col_in
         || ' = :val 
        WHERE department_id = :dept and :val IS NOT NULL';
BEGIN
   EXECUTE IMMEDIATE l_update USING val_in, department_id_in, val_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/

BEGIN
   DBMS_OUTPUT.put_line (
      'Before '
      || table_count (USER
                    , 'employees_m2'
                    , 'department_id = 50 AND salary = 10000'));
   updnumval ('salary', 50, 10000);
   DBMS_OUTPUT.put_line (
      'After '
      || table_count (USER
                    , 'employees_m2'
                    , 'department_id = 50 AND salary = 10000'));
   ROLLBACK;
END;
/

/* Cannot reference PL/SQL types */

CREATE OR REPLACE PROCEDURE updnumval (rec_in             IN employees_m2%ROWTYPE
                                     , department_id_in   IN PLS_INTEGER
                                     , val_in             IN NUMBER)
IS
   l_update   VARCHAR2 (1000)
      := 'UPDATE employees_m2 SET ROW = :rowdata
              WHERE department_id = :dept and :val IS NOT NULL';
BEGIN
   EXECUTE IMMEDIATE l_update
      USING rec_in
          , department_id_in
          , val_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/