DROP TABLE alert_users
/

DROP TABLE alert_follows
/

DROP TABLE alert_log
/

DROP TABLE alert_runs
/

DROP TABLE alert_actions
/

DROP TABLE alerts
/

DROP TABLE alert_config
/

CREATE TABLE alert_config
(
   job_wakeup_minutes   INTEGER,
   keep_checking        VARCHAR2 (1),
   send_email_proc_name VARCHAR2 (300),
   apex_workspace_name  VARCHAR2(100)
)
/

BEGIN
   INSERT INTO alert_config
        VALUES (5, 'Y', 'alert_email_mgr.send_email', 'MIKE');

   COMMIT;
END;
/

CREATE TABLE alerts (
   id                     NUMBER NOT NULL CONSTRAINT alerts_id_pk PRIMARY KEY,
   alert_name             VARCHAR2 (100) CONSTRAINT alerts_alert_name_unq UNIQUE,
   description            VARCHAR2 (4000),
   trigger_code           VARCHAR2 (4000),
   trigger_type           VARCHAR2 (100),
   check_frequency        VARCHAR2 (1000),
   recurrence_frequency   VARCHAR2 (1000),
   ignore_failures        VARCHAR2 (1),
   is_active              VARCHAR2 (1),
   last_run_id            NUMBER,
   created                DATE NOT NULL,
   created_by             VARCHAR2 (255) NOT NULL,
   updated                DATE NOT NULL,
   updated_by             VARCHAR2 (255) NOT NULL,
   CONSTRAINT trigger_type_choices CHECK
      (trigger_type IN ('ONEROW',
                        'NOROWS',
                        'MULTROWS',
                        'RETURNTRUE',
                        'RETURNFALSE')),
   CONSTRAINT ignore_failures_yn CHECK (ignore_failures IN ('Y', 'N')),
   CONSTRAINT active_yn CHECK (is_active IN ('Y', 'N'))
)
/

COMMENT ON COLUMN alerts.trigger_code IS
   'Code executed to determine if an alert should be triggered'
/

COMMENT ON COLUMN alerts.trigger_type IS
   'Type of code executed to determine if an alert should be triggered. See trigger_type_choices constraint.'
/

COMMENT ON COLUMN alerts.check_frequency IS
   'How often to check to see if alert is triggered. H for hourly, D for daily, # for number of minutes since last triggering.'
/

COMMENT ON COLUMN alerts.recurrence_frequency IS
   'How often to trigger an alert since last triggering. H for hourly, D for daily, # for number of minutes since last triggering.'
/

COMMENT ON COLUMN alerts.ignore_failures IS
   'If Y then failure of last triggered action does not count for recurrence frequency.'
/

COMMENT ON COLUMN alerts.is_active IS
   'If Y then include in alert checking, otherwise ignore.'
/

COMMENT ON COLUMN alerts.last_run_id IS
   'Primary key of row in alert_runs table which is the last run for this alert.'
/

CREATE TABLE alert_actions (
   id                NUMBER NOT NULL CONSTRAINT alert_action_id_pk PRIMARY KEY,
   alert_id          NUMBER
                        CONSTRAINT alert_action_alert_id_fk
                            REFERENCES alerts ON DELETE CASCADE,
   action_sequence   INTEGER,
   action_type       VARCHAR2 (100) NOT NULL,
   action_code       VARCHAR2 (4000),
   send_email_to     VARCHAR2 (4000),
   email_subject     VARCHAR2 (4000),
   email_body        CLOB,
   email_query       VARCHAR2 (4000),
   created           DATE NOT NULL,
   created_by        VARCHAR2 (255) NOT NULL,
   updated           DATE NOT NULL,
   updated_by        VARCHAR2 (255) NOT NULL,
   CONSTRAINT action_type_choices_cc CHECK
      (action_type IN ('EMAIL',
                       'SMS',
                       'PROC',
                       'REST'))
)
/

COMMENT ON COLUMN alert_actions.action_type IS
   'The type of action taken. See action_type_choices for options.'
/

COMMENT ON COLUMN alert_actions.action_code IS 'Code used to execute action.'
/

COMMENT ON COLUMN alert_actions.send_email_to IS
   'Comma-delimited list of email addresses if action is to send an email'
/

COMMENT ON COLUMN alert_actions.email_query IS
   'SELECT subject, body - used to send multiple emails'
/

CREATE TABLE alert_runs (
   id                       NUMBER NOT NULL CONSTRAINT alert_runs_id_pk PRIMARY KEY,
   alert_id                 NUMBER
                               CONSTRAINT alert_runs_alert_id_fk
                                   REFERENCES alerts ON DELETE CASCADE,
   started_at               DATE,
   completed_at             DATE,
   completed_successfully   VARCHAR2 (1),
   results                  VARCHAR2 (4000),
   errors                   VARCHAR2 (4000),
   created                  DATE NOT NULL,
   created_by               VARCHAR2 (255) NOT NULL,
   updated                  DATE NOT NULL,
   updated_by               VARCHAR2 (255) NOT NULL,
   CONSTRAINT success_yn CHECK (completed_successfully IN ('Y', 'N'))
)
/

CREATE TABLE alert_log
(
   id           NUMBER NOT NULL CONSTRAINT alert_log_id_pk PRIMARY KEY,
   info         VARCHAR2 (4000),
   created      DATE NOT NULL,
   created_by   VARCHAR2 (255) NOT NULL,
   updated      DATE NOT NULL,
   updated_by   VARCHAR2 (255) NOT NULL
)
/

CREATE TABLE alert_users
(
   id           NUMBER NOT NULL CONSTRAINT alert_user_id_pk PRIMARY KEY,
   email_address VARCHAR2 (4000),
   created      DATE NOT NULL,
   created_by   VARCHAR2 (255) NOT NULL,
   updated      DATE NOT NULL,
   updated_by   VARCHAR2 (255) NOT NULL
)
/

CREATE TABLE alert_follows
(
   id           NUMBER NOT NULL CONSTRAINT alert_follow_id_pk PRIMARY KEY,
   user_id      NUMBER NOT NULL,
   alert_id     NUMBER NOT NULL,
   created      DATE NOT NULL,
   created_by   VARCHAR2 (255) NOT NULL,
   updated      DATE NOT NULL,
   updated_by   VARCHAR2 (255) NOT NULL
)
/

CREATE OR REPLACE TRIGGER alerts_biu
   BEFORE INSERT OR UPDATE
   ON alerts
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   :new.alert_name := UPPER (:new.alert_name);
   :new.trigger_type := UPPER (:new.trigger_type);
END alerts_biu;
/

CREATE OR REPLACE TRIGGER alert_actions_biu
   BEFORE INSERT OR UPDATE
   ON alert_actions
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   :new.action_type := UPPER (:new.action_type);
END;
/

CREATE OR REPLACE TRIGGER alert_runs_biu
   BEFORE INSERT OR UPDATE
   ON alert_runs
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
END alert_runs_biu;
/

CREATE OR REPLACE TRIGGER alert_log_biu
   BEFORE INSERT OR UPDATE
   ON alert_log
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
END alert_log_biu;
/

CREATE OR REPLACE TRIGGER alert_users_biu
   BEFORE INSERT OR UPDATE
   ON alert_users
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   
   :new.email_address := UPPER (:new.email_address);
END;
/

CREATE OR REPLACE TRIGGER alert_follows_biu
   BEFORE INSERT OR UPDATE
   ON alert_follows
   FOR EACH ROW
BEGIN
   IF :new.id IS NULL
   THEN
      :new.id := TO_NUMBER (SYS_GUID (), 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX');
   END IF;

   IF INSERTING
   THEN
      :new.created := SYSDATE;
      :new.created_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
   END IF;

   :new.updated := SYSDATE;
   :new.updated_by := NVL (SYS_CONTEXT ('APEX$SESSION', 'APP_USER'), USER);
END;
/

CREATE INDEX alert_runs_i1
   ON alert_runs (alert_id)
/

CREATE UNIQUE INDEX alert_users_ui1
   ON alert_users (UPPER (email_address))
/

CREATE UNIQUE INDEX alert_follows_ui1
   ON alert_follows (alert_id, user_id)
/

ALTER TABLE alert_follows ADD (
     CONSTRAINT fk_alert_follows_u
     FOREIGN KEY (user_id)
     REFERENCES alert_users (id)
     ENABLE VALIDATE)
/

ALTER TABLE alert_follows ADD (
     CONSTRAINT fk_alert_follows_a
     FOREIGN KEY (alert_id)
     REFERENCES alerts (id)
     ENABLE VALIDATE)
/